from mindee.fields import Field


class Orientation(Field):
    def __init__(
            self,
            orientation_prediction,
            value_key="degrees",
            reconstructed=False,
            page_n=None
    ):
        """
        :param orientation_prediction: Orientation prediction object from HTTP response
        :param value_key: Key to use in the orientation_prediction dict
        :param reconstructed: Bool for reconstructed object (not extracted in the API)
        :param page_n: Page number for multi pages pdf
        """
        super(Orientation, self).__init__(
            orientation_prediction,
            value_key=value_key,
            reconstructed=reconstructed,
            page_n=page_n
        )

        try:
            self.value = int(orientation_prediction[value_key])
            if self.value not in [0, 90, 180, 270]:
                self.value = 0
        except:
            self.value = 0
            self.probability = 0.



from mindee.fields.orientation import Orientation


def test_constructor():
    field_dict = {
        'degrees': 90,
        'probability': 0.1,
    }
    orientation = Orientation(field_dict)
    assert orientation.value == 90


def test_not_number():
    field_dict = {
        'degrees': "aze",
        'probability': 0.1,
    }
    orientation = Orientation(field_dict)
    assert orientation.value == 0


def test_not_90():
    field_dict = {
        'degrees': 255,
        'probability': 0.1,
    }
    orientation = Orientation(field_dict)
    assert orientation.value == 0



from mindee.fields import Field


class Tax(Field):
    def __init__(
            self,
            tax_prediction,
            value_key="amount",
            rate_key="rate",
            code_key="code",
            reconstructed=False,
            page_n=None
    ):
        """
        :param tax_prediction: Tax prediction object from HTTP response
        :param value_key: Key to use in the tax_prediction dict
        :param rate_key: Key to use for getting the Tax rate in the tax_prediction dict
        :param code_key: Key to use for getting the Tax code in the tax_prediction dict
        :param reconstructed: Bool for reconstructed object (not extracted in the API)
        :param page_n: Page number for multi pages pdf
        """
        super(Tax, self).__init__(
            tax_prediction,
            value_key=value_key,
            reconstructed=reconstructed,
            page_n=page_n
        )

        try:
            self.rate = float(tax_prediction[rate_key])
        except:
            self.rate = None

        try:
            self.code = str(tax_prediction[code_key])
            if self.code == "N/A":
                self.code = None
        except:
            self.code = None

        try:
            self.value = float(tax_prediction[value_key])
        except:
            self.value = None
            self.probability = 0.

    def __str__(self):
        tax_str = ""
        if self.value is not None:
            tax_str += str(self.value)
        else:
            tax_str += "_"

        if self.rate is not None:
            tax_str += "; "+str(self.rate)+"%"
        else:
            tax_str += "; _"

        if self.code is not None:
            tax_str += "; "+str(self.code)
        else:
            tax_str += "; _"

        return tax_str





from mindee.documents import Document
from mindee.fields import Field
from mindee.http import request
import os


class CarPlate(Document):
    def __init__(
            self,
            api_prediction=None,
            input_file=None,
            license_plates=None,
            page_n=0
    ):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param input_file: Input object
        :param license_plates: List of license plates values for creating CarPlate object from scratch
        :param page_n: Page number for multi pages pdf input
        """
        self.type = "CarPlate"
        self.license_plates = []

        if api_prediction is not None:
            self.build_from_api_prediction(api_prediction, page_n=page_n)
        else:
            if license_plates is not None:
                self.license_plates = [
                    Field({"value": l}, value_key="value", page_n=page_n) for l in license_plates
                ]

        # Invoke Document constructor
        super(CarPlate, self).__init__(input_file)

        # Run checks
        self._checklist()

        # Reconstruct extra fields
        self._reconstruct()

    def build_from_api_prediction(self, api_prediction, page_n=0):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param page_n: Page number for multi pages pdf input
        :return: (void) set the object attributes with api prediction values
        """
        self.license_plates = [
            Field(license_plate, page_n=page_n) for license_plate in api_prediction["license_plates"]
        ]

    def __str__(self):
        return "-----Car plate data-----\n" \
               "Filename: %s\n" \
               "Plate: %s\n" \
               "----------------------" % \
               (self.filename, " ".join([l.value for l in self.license_plates]))

    @staticmethod
    def compare(license_plate=None, ground_truth=None):
        """
        :param license_plate: CarPlate object to compare
        :param ground_truth: Ground truth CarPlate object
        :return: Accuracy and precisions metrics
        """
        assert license_plate is not None
        assert ground_truth is not None
        assert isinstance(license_plate, CarPlate)
        assert isinstance(ground_truth, CarPlate)

        metrics = {}

        # Compute Accuracy metrics
        metrics.update(CarPlate.compute_accuracy(license_plate, ground_truth))

        # Compute precision metrics
        metrics.update(CarPlate.compute_precision(license_plate, ground_truth))

        return metrics

    @staticmethod
    def request(input_file, base_url, license_plates_token=None, version="1"):
        """
        Make request to license_plates endpoint
        :param input_file: Input object
        :param base_url: API base URL
        :param license_plates_token: License plate API token
        :param version: API version
        """
        url = os.path.join(base_url, "license_plates", "v" + version, "predict")
        return request(url, input_file, license_plates_token)

    def _checklist(self):
        """
        Call check methods
        """
        pass

    def _reconstruct(self):
        """
        Call fields reconstruction methods
        """
        pass

    @staticmethod
    def compute_accuracy(license_plate, ground_truth):
        """
        :param license_plate: CarPlate object to compare
        :param ground_truth: Ground truth CarPlate object
        :return: Accuracy metrics
        """
        return {
            "__acc__license_plates": Field.compare_arrays(
                license_plate.license_plates, ground_truth.license_plates)
        }

    @staticmethod
    def compute_precision(license_plate, ground_truth):
        """
        :param license_plate: CarPlate object to compare
        :param ground_truth: Ground truth CarPlate object
        :return: Precisions metrics
        """
        precisions = {}

        if len(license_plate.license_plates) == 0:
            precisions["__pre__license_plates"] = None
        else:
            precisions["__pre__license_plates"] = Field.compare_arrays(
                license_plate.license_plates, ground_truth.license_plates)

        return precisions



from mindee.fields import Field


def test_constructor():
    field_dict = {
        'value': "test",
        'probability': 0.1,
        'segmentation': {
            "bounding_box": [
                [0.016, 0.707],
                [0.414, 0.707],
                [0.414, 0.831],
                [0.016, 0.831]
            ]
        }
    }
    field = Field(field_dict)
    assert field.value == "test"
    assert field.probability == 0.1
    assert len(field.bbox) > 0


def test_constructor_no_segmentation():
    field_dict = {
        'value': "test",
        'probability': 0.1
    }
    field = Field(field_dict)
    assert len(field.bbox) == 0


def test_equality():
    field_dict_1 = {
        'value': "test",
        'probability': 0.1
    }
    field_dict_2 = {
        'value': "other",
        'probability': 0.1
    }
    field_1 = Field(field_dict_1)
    field_2 = Field(field_dict_2)
    assert field_1 == field_1
    assert field_1 != field_2


def test_constructor_na():
    field_dict = {
        'value': "N/A",
        'probability': 0.1
    }
    field = Field(field_dict)
    assert field.value is None


def test_no_probability():
    field_dict = {
        'value': "N/A",
    }
    field = Field(field_dict)
    assert field.probability == 0.


def test_array_probability():
    fields = [Field({"value": None, "probability": 0.1}), Field({"value": None, "probability": 0.8})]
    assert Field.array_probability(fields) == 0.8*0.1
    fields = [Field({"value": None, "probability": 0.1}), Field({"value": None, "probability": None})]
    assert Field.array_probability(fields) == 0.


def test_array_sum():
    fields = [Field({"value": 1, "probability": 0.1}), Field({"value": 2, "probability": 0.8})]
    assert Field.array_sum(fields) == 3
    fields = [Field({"value": None, "probability": 0.1}), Field({"value": 4, "probability": 0.8})]
    assert Field.array_sum(fields) == 0.



from mindee.fields import Field


class Amount(Field):
    def __init__(
            self,
            amount_prediction,
            value_key="amount",
            reconstructed=False,
            page_n=None
    ):
        """
        :param amount_prediction: Amount prediction object from HTTP response
        :param value_key: Key to use in the amount_prediction dict
        :param reconstructed: Bool for reconstructed object (not extracted in the API)
        :param page_n: Page number for multi pages pdf
        """
        super(Amount, self).__init__(
            amount_prediction,
            value_key=value_key,
            reconstructed=reconstructed,
            page_n=page_n
        )

        try:
            self.value = round(float(amount_prediction[value_key]), 3)
        except:
            self.value = None
            self.probability = 0.



import fitz
import io
import os
from base64 import decodebytes
from mimetypes import guess_type


class Inputs(object):
    def __init__(
            self,
            file,
            input_type="path",
            filename=None,
            cut_pdf=True
    ):
        """
        :param file: Either path or base64 string, or stream
        :param input_type: Specify the type of input fed into the Input
        :param filename: File name of the input
        :param cut_pdf: Automatically reconstruct pdf with more than 4 pages
        """
        self.allowed_extensions = ["image/png", "image/jpg", "image/jpeg", "image/webp", "application/pdf"]
        assert input_type in ["base64", "path", "stream", "dummy"]

        if input_type == "base64":
            # Only for images
            self.file_object = Inputs.b64_to_stream(file)
            self.input_type = input_type
            self.filename = filename
            self.filepath = None
            self.file_extension = 'image/jpg'
        elif input_type == "stream":
            # Case input is a file object
            self.file_object = file
            self.input_type = input_type
            self.filename = os.path.basename(file.name)
            self.filepath = file.name
            self.file_extension = guess_type(file.name)[0]
        elif input_type == "path":
            # Case input is a path
            self.file_object = open(file, "rb")
            self.input_type = input_type
            self.filename = os.path.basename(file)
            self.filepath = file
            self.file_extension = guess_type(file)[0]

        if input_type == "dummy":
            self.file_object = ""
            self.input_type = ""
            self.filename = ""
            self.filepath = ""
            self.file_extension = ""
        elif self.file_extension not in self.allowed_extensions:
            raise Exception("File type not allowed, must be in {%s}" % ", ".join(self.allowed_extensions))

        if self.file_extension == "application/pdf" and cut_pdf is True:
            n_pages = self.count_pdf_pages()
            if n_pages > 3:
                self.merge_pdf_pages({0, n_pages - 2, n_pages - 1})

    @staticmethod
    def load(input_type, filename, filepath, file_extension):
        """
        :param input_type: Specify the type of input fed into the Input
        :param filename: File name of the input
        :param filepath: Original file path of the Input file
        :param file_extension: Extension of the file
        :return: Dummy Input object to use when restoring Response from json file
        """
        file_input = Inputs(filename, input_type="dummy")
        file_input.input_type = input_type
        file_input.filepath = filepath
        file_input.file_extension = file_extension
        return file_input

    @staticmethod
    def b64_to_stream(b64_string):
        """
        :param b64_string: image base 64 string
        :return: stream from base64
        """
        bytes_object = decodebytes(b64_string.encode("utf-8"))
        return io.BytesIO(bytes_object)

    def count_pdf_pages(self):
        """
        :return: Number of pages in the Input file for pdfs
        """
        self.file_object.seek(0)
        src = fitz.open(
            stream=self.file_object.read(),
            filetype=self.file_extension,
            filename=self.filename
        )
        return len(src)

    def merge_pdf_pages(self, pages_number):
        """
        :param pages_number: List of pages number to use for merging in the original pdf
        :return: (void) Set the Input.file with the reconstructed pdf stream
        """
        self.file_object.seek(0)
        src = fitz.open(
            stream=self.file_object.read(),
            filetype="pdf"
        )
        doc = fitz.open()
        pdf_pages = [src[n] for n in pages_number]
        for spage in pdf_pages:
            width = spage.MediaBoxSize[0]
            height = spage.MediaBoxSize[1]
            r = fitz.Rect(0, 0, width, height)
            page = doc.newPage(-1, width=width, height=height)
            page.showPDFpage(r, src, spage.number)
        self.file_object.close()
        self.file_object = io.BytesIO(doc.write())




from mindee.documents import Document
from mindee.fields import Field
from mindee.fields.date import Date
from mindee.fields.amount import Amount
from mindee.fields.locale import Locale
from mindee.fields.orientation import Orientation
from mindee.fields.tax import Tax
from mindee.http import request
from mindee.benchmark import Benchmark
import os


class Invoice(Document):
    def __init__(
            self,
            api_prediction=None,
            input_file=None,
            locale=None,
            total_incl=None,
            total_excl=None,
            invoice_date=None,
            invoice_number=None,
            due_date=None,
            taxes=None,
            supplier=None,
            payment_details=None,
            company_number=None,
            vat_number=None,
            orientation=None,
            total_tax=None,
            page_n=0
    ):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param input_file: Input object
        :param locale: locale value for creating Invoice object from scratch
        :param total_incl: total_incl value for creating Invoice object from scratch
        :param total_excl: total_excl value for creating Invoice object from scratch
        :param invoice_date: invoice_date value for creating Invoice object from scratch
        :param invoice_number: invoice_number value for creating Invoice object from scratch
        :param due_date: due_date value for creating Invoice object from scratch
        :param taxes: taxes value for creating Invoice object from scratch
        :param supplier: supplier value for creating Invoice object from scratch
        :param payment_details: payment_details value for creating Invoice object from scratch
        :param company_number: company_number value for creating Invoice object from scratch
        :param vat_number: vat_number value for creating Invoice object from scratch
        :param orientation: orientation value for creating Invoice object from scratch
        :param total_tax: total_tax value for creating Invoice object from scratch
        :param page_n: Page number for multi pages pdf input
        """
        self.type = "Invoice"
        self.locale = None
        self.total_incl = None
        self.total_excl = None
        self.invoice_date = None
        self.invoice_number = None
        self.due_date = None
        self.taxes = []
        self.supplier = None
        self.payment_details = None
        self.company_number = None
        self.vat_number = None
        self.orientation = None
        self.total_tax = None

        if api_prediction is not None:
            self.build_from_api_prediction(api_prediction, page_n=page_n)
        else:
            self.locale = Locale({"value": locale}, value_key="value", page_n=page_n)
            self.total_incl = Amount({"value": total_incl}, value_key="value", page_n=page_n)
            self.date = Date({"value": invoice_date}, value_key="value", page_n=page_n)
            self.invoice_date = Date({"value": invoice_date}, value_key="value", page_n=page_n)
            self.due_date = Date({"value": due_date}, value_key="value", page_n=page_n)
            self.supplier = Field({"value": supplier}, value_key="value", page_n=page_n)
            if taxes is not None:
                self.taxes = [
                    Tax({"value": t[0], "rate": t[1]}, page_n=page_n, value_key="value", rate_key="rate")
                    for t in taxes]
            self.orientation = Orientation({"value": orientation}, value_key="value", page_n=page_n)
            self.total_tax = Amount({"value": total_tax}, value_key="value", page_n=page_n)
            self.total_excl = Amount({"value": total_excl}, value_key="value", page_n=page_n)
            self.invoice_number = Field({"value": invoice_number}, value_key="value", page_n=page_n)
            self.payment_details = Field({"value": payment_details}, value_key="value", page_n=page_n)
            self.company_number = Field({"value": company_number}, value_key="value", page_n=page_n)
            self.vat_number = Field({"value": vat_number}, value_key="value", page_n=page_n)

        # Invoke Document constructor
        super(Invoice, self).__init__(input_file)

        # Run checks
        self._checklist()

        # Reconstruct extra fields
        self._reconstruct()

    def build_from_api_prediction(self, api_prediction, page_n=0):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param page_n: Page number for multi pages pdf input
        :return: (void) set the object attributes with api prediction values
        """
        self.locale = Locale(api_prediction["locale"], value_key="language", page_n=page_n)
        self.total_incl = Amount(api_prediction["total_incl"], page_n=page_n)
        self.total_excl = Amount(api_prediction["total_excl"], page_n=page_n)
        self.invoice_date = Date(api_prediction["invoice_date"], page_n=page_n)
        self.invoice_number = Field(api_prediction["invoice_number"], page_n=page_n)
        self.due_date = Date(api_prediction["due_date"], page_n=page_n)
        self.taxes = [Tax(tax_prediction, page_n=page_n) for tax_prediction in api_prediction["taxes"]]
        self.supplier = Field(api_prediction["supplier"], page_n=page_n)
        self.payment_details = Field(api_prediction["payment_details"], value_key="iban", page_n=page_n)
        self.vat_number = Field(api_prediction["tax_id"], page_n=page_n)
        self.company_number = Field(api_prediction["company_number"], extra_fields={"type"}, page_n=page_n)
        self.orientation = Orientation(api_prediction["orientation"], page_n=page_n)
        self.total_tax = Amount({"value": None, "probability": 0.}, value_key="value", page_n=page_n)

    def __str__(self):
        return "-----Invoice data-----\n" \
               "Filename: %s \n" \
               "Invoice number: %s \n" \
               "Total amount including taxes: %s \n" \
               "Total amount excluding taxes: %s \n" \
               "Invoice date: %s\n" \
               "Supplier name: %s\n" \
               "Taxes: %s\n" \
               "Total taxes: %s\n" \
               "----------------------" % \
               (
                   self.filename,
                   self.invoice_number.value,
                   self.total_incl.value,
                   self.total_excl.value,
                   self.invoice_date.value,
                   self.supplier.value,
                   ",".join([str(t.value) + " " + str(t.rate) + "%" for t in self.taxes]),
                   self.total_tax.value
               )

    @staticmethod
    def compare(invoice=None, ground_truth=None):
        """
        :param invoice: Invoice object to compare
        :param ground_truth: Ground truth Invoice object
        :return: Accuracy and precisions metrics
        """
        assert invoice is not None
        assert ground_truth is not None
        assert isinstance(invoice, Invoice)
        assert isinstance(ground_truth, Invoice)

        metrics = {}

        # Compute Accuracy metrics
        metrics.update(Invoice.compute_accuracy(invoice, ground_truth))

        # Compute precision metrics
        metrics.update(Invoice.compute_precision(invoice, ground_truth))

        return metrics

    @staticmethod
    def request(
            input_file,
            base_url,
            invoice_token=None,
            version="1",
            include_words=False
    ):
        """
        Make request to invoices endpoint
        :param input_file: Input object
        :param base_url: API base URL
        :param invoice_token: Invoices API token
        :param include_words: Include Mindee vision words in http_response
        :param version: API version
        """
        url = os.path.join(base_url, "invoices", "v" + version, "predict")
        return request(url, input_file, invoice_token, include_words)

    def _reconstruct(self):
        """
        Call fields reconstruction methods
        """
        self.__reconstruct_total_tax()
        self.__reconstruct_total_excl_from_tcc_and_taxes()
        self.__reconstruct_total_incl_from_taxes_plus_excl()

    def _checklist(self):
        """
        Call check methods
        """
        self.checklist = {
            "taxes_match_total_incl": self.__taxes_match_total_incl(),
            "taxes_match_total_excl": self.__taxes_match_total_excl(),
            "taxes_plus_total_excl_match_total_incl": self.__taxes_plus_total_excl_match_total_incl()
        }

    # Checks
    def __taxes_match_total_incl(self):
        """
        Check invoice rule of matching between taxes and total_incl
        :return: True if rule matches, False otherwise
        """
        # Check taxes and total_incl exist
        if len(self.taxes) == 0 or self.total_incl.value is None:
            return False

        # Reconstruct total_incl from taxes
        total_vat = 0
        reconstructed_total = 0
        for tax in self.taxes:
            total_vat += tax.value
            reconstructed_total += tax.value + 100 * tax.value / tax.rate

        # Sanity check
        if total_vat <= 0:
            return False

        # Crate epsilon
        eps = 1 / (100 * total_vat)

        if self.total_incl.value * (1 - eps) - 0.02 <= reconstructed_total <= self.total_incl.value * (
                1 + eps) + 0.02:
            for tax in self.taxes:
                tax.probability = 1
            self.total_tax.probability = 1.
            self.total_incl.probability = 1.
            return True
        else:
            return False

    def __taxes_match_total_excl(self):
        """
        Check invoice rule of matching between taxes and total_excl
        :return: True if rule matches, False otherwise
        """
        # Check taxes and total excl exist
        if len(self.taxes) == 0 or self.total_excl.value is None:
            return False

        # Reconstruct total excl from taxes
        total_vat = 0
        reconstructed_total = 0
        for tax in self.taxes:
            total_vat += tax.value
            reconstructed_total += 100 * tax.value / tax.rate

        # Sanity check
        if total_vat <= 0:
            return False

        # Crate epsilon
        eps = 1 / (100 * total_vat)

        # Check that reconstructed total excl matches total excl
        if self.total_excl.value * (1 - eps) - 0.02 <= reconstructed_total <= self.total_excl.value * (
                1 + eps) + 0.02:
            for tax in self.taxes:
                tax.probability = 1
            self.total_tax.probability = 1.
            self.total_excl.probability = 1.
            return True
        else:
            return False

    def __taxes_plus_total_excl_match_total_incl(self):
        """
        Check invoice rule of matching : sum(taxes) + total_excluding_taxes = total_including_taxes
        :return: True if rule matches, False otherwise
        """
        # Check total_tax, total excl and total incl exist
        if self.total_excl.value is None or len(self.taxes) == 0 or self.total_incl.value is None:
            return False

        # Reconstruct total_incl
        total_vat = 0
        for tax in self.taxes:
            total_vat += tax.value
        reconstructed_total = total_vat + self.total_excl.value

        # Sanity check
        if total_vat <= 0:
            return False

        # Check that reconstructed total incl matches total excl + taxes sum
        if self.total_incl.value - 0.01 <= reconstructed_total <= self.total_incl.value + 0.01:
            for tax in self.taxes:
                tax.probability = 1
            self.total_tax.probability = 1.
            self.total_excl.probability = 1.
            self.total_incl.probability = 1.
            return True
        else:
            return False

    # Reconstruct
    def __reconstruct_total_incl_from_taxes_plus_excl(self):
        """
        Set self.total_incl with Amount object
        The total_incl Amount value is the sum of total_excl and sum of taxes
        The total_incl Amount probability is the product of self.taxes probabilities multiplied by total_excl probability
        """
        # Check total_tax, total excl exist and total incl is not set
        if self.total_excl.value is None or len(self.taxes) == 0 or self.total_incl.value is not None:
            pass
        else:
            total_incl = {
                "value": sum([tax.value if tax.value is not None else 0 for tax in self.taxes]) + self.total_excl.value,
                "probability": Field.array_probability(self.taxes) * self.total_excl.probability
            }
            self.total_incl = Amount(total_incl, value_key="value", reconstructed=True)

    def __reconstruct_total_excl_from_tcc_and_taxes(self):
        """
        Set self.total_excl with Amount object
        The total_excl Amount value is the difference between total_incl and sum of taxes
        The total_excl Amount probability is the product of self.taxes probabilities multiplied by total_incl probability
        """
        # Check total_tax, total excl and total incl exist
        if self.total_incl.value is None or len(self.taxes) == 0 or self.total_excl.value is not None:
            pass
        else:
            total_excl = {
                "value": self.total_incl.value - sum([tax.value if tax.value is not None else 0 for tax in self.taxes]),
                "probability": Field.array_probability(self.taxes) * self.total_incl.probability
            }
            self.total_excl = Amount(total_excl, value_key="value", reconstructed=True)

    def __reconstruct_total_tax(self):
        """
        Set self.total_tax with Amount object
        The total_tax Amount value is the sum of all self.taxes value
        The total_tax Amount probability is the product of self.taxes probabilities
        """
        if len(self.taxes):
            total_tax = {
                "value": sum([tax.value if tax.value is not None else 0 for tax in self.taxes]),
                "probability": Field.array_probability(self.taxes)
            }
            if total_tax["value"] > 0:
                self.total_tax = Amount(total_tax, value_key="value", reconstructed=True)

    @staticmethod
    def compute_accuracy(invoice, ground_truth):
        """
        :param invoice: Invoice object to compare
        :param ground_truth: Ground truth Invoice object
        :return: Accuracy metrics
        """
        return {
            "__acc__total_incl": ground_truth.total_incl == invoice.total_incl,
            "__acc__total_excl": ground_truth.total_excl == invoice.total_excl,
            "__acc__invoice_date": ground_truth.invoice_date == invoice.invoice_date,
            "__acc__invoice_number": ground_truth.invoice_number == invoice.invoice_number,
            "__acc__due_date": ground_truth.due_date == invoice.due_date,
            "__acc__total_tax": ground_truth.total_tax == invoice.total_tax,
            "__acc__taxes": Tax.compare_arrays(invoice.taxes, ground_truth.taxes),
        }

    @staticmethod
    def compute_precision(invoice, ground_truth):
        """
        :param invoice: Invoice object to compare
        :param ground_truth: Ground truth Invoice object
        :return: Precision metrics
        """
        precisions = {
            "__pre__total_incl": Benchmark.scalar_precision_score(
                invoice.total_incl, ground_truth.total_incl),
            "__pre__total_excl": Benchmark.scalar_precision_score(
                invoice.total_excl, ground_truth.total_excl),
            "__pre__invoice_date": Benchmark.scalar_precision_score(
                invoice.invoice_date, ground_truth.invoice_date),
            "__pre__invoice_number": Benchmark.scalar_precision_score(
                invoice.invoice_number, ground_truth.invoice_number),
            "__pre__due_date": Benchmark.scalar_precision_score(
                invoice.due_date, ground_truth.due_date),
            "__pre__total_tax": Benchmark.scalar_precision_score(
                invoice.total_tax, ground_truth.total_tax)}

        if len(invoice.taxes) == 0:
            precisions["__pre__taxes"] = None
        else:
            precisions["__pre__taxes"] = Tax.compare_arrays(invoice.taxes, ground_truth.taxes)

        return precisions



import json

import pytest

from mindee import CarPlate


@pytest.fixture
def car_plate_object():
    json_repsonse = json.load(open("./tests/data/license_plates/v1/plate.json"))
    return CarPlate(json_repsonse["predictions"][0])


@pytest.fixture
def car_plate_object_from_scratch():
    return CarPlate(
        license_plates=["abcdef", "klimopo"]
    )


@pytest.fixture
def car_plate_object_all_na():
    json_repsonse = json.load(open("./tests/data/license_plates/v1/plate_all_na.json"))
    return CarPlate(json_repsonse["predictions"][0])


def test_constructor(car_plate_object):
    assert car_plate_object.license_plates[0].value == "7EQE707"


def test__str__(car_plate_object):
    assert type(car_plate_object.__str__()) == str


def test_all_na(car_plate_object_all_na):
    assert len(car_plate_object_all_na.license_plates) == 0


def test_compare_1(car_plate_object):
    # Compare same object must return all True
    benchmark = CarPlate.compare(car_plate_object, car_plate_object)
    for value in benchmark.values():
        assert value is True


def test_compare_2(car_plate_object, car_plate_object_all_na):
    # Compare full object and empty object
    benchmark = CarPlate.compare(car_plate_object, car_plate_object_all_na)
    for value in benchmark.values():
        assert value is False


def test_compare_3(car_plate_object_from_scratch):
    # Compare car plates from class
    benchmark = CarPlate.compare(
        car_plate_object_from_scratch,
        car_plate_object_from_scratch
    )
    for key in benchmark.keys():
        if "__acc__" in key:
            assert benchmark[key] is True


def test_compare_4(car_plate_object_from_scratch):
    # Compare car plates from class with empty taxes
    car_plate_object_from_scratch.license_plates = []
    benchmark = CarPlate.compare(
        car_plate_object_from_scratch,
        car_plate_object_from_scratch
    )
    for key in benchmark.keys():
        if "__acc__" in key:
            assert benchmark[key] is True
        elif "__pre__" in key:
            assert benchmark[key] in [True, None]


def test_empty_object_works():
    car_plate = CarPlate()
    assert len(car_plate.license_plates) == 0



from mindee.benchmark import Benchmark
from mindee.documents import Document
from mindee.fields import Field
from mindee.fields.date import Date
from mindee.fields.amount import Amount
from mindee.fields.locale import Locale
from mindee.fields.orientation import Orientation
from mindee.fields.tax import Tax
from mindee.http import request
import os


class Receipt(Document):
    def __init__(
            self,
            api_prediction=None,
            input_file=None,
            locale=None,
            total_incl=None,
            date=None,
            category=None,
            merchant_name=None,
            time=None,
            taxes=None,
            orientation=None,
            total_tax=None,
            total_excl=None,
            page_n=0
    ):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param input_file: Input object
        :param locale: locale value for creating Receipt object from scratch
        :param total_incl: total_incl value for creating Receipt object from scratch
        :param date: date value for creating Receipt object from scratch
        :param category: category value for creating Receipt object from scratch
        :param merchant_name: merchant_name value for creating Receipt object from scratch
        :param time: time value for creating Receipt object from scratch
        :param taxes: taxes value for creating Receipt object from scratch
        :param orientation: orientation value for creating Receipt object from scratch
        :param total_tax: total_tax value for creating Receipt object from scratch
        :param total_excl: total_excl value for creating Receipt object from scratch
        :param page_n: Page number for multi pages pdf input
        """
        self.type = "Receipt"
        self.locale = None
        self.total_incl = None
        self.date = None
        self.category = None
        self.merchant_name = None
        self.time = None
        self.taxes = []
        self.orientation = None
        self.total_tax = None
        self.total_excl = None

        if api_prediction is not None:
            self.build_from_api_prediction(api_prediction, page_n=page_n)
        else:
            self.locale = Locale({"value": locale}, value_key="value", page_n=page_n)
            self.total_incl = Amount({"value": total_incl}, value_key="value", page_n=page_n)
            self.date = Date({"value": date}, value_key="value", page_n=page_n)
            self.category = Field({"value": category}, value_key="value", page_n=page_n)
            self.merchant_name = Field({"value": merchant_name}, value_key="value", page_n=page_n)
            self.time = Field({"value": time}, value_key="value", page_n=page_n)
            if taxes is not None:
                self.taxes = [
                    Tax({"value": t[0], "rate": t[1]}, page_n=page_n, value_key="value", rate_key="rate")
                    for t in taxes]
            self.orientation = Orientation({"value": orientation}, value_key="value", page_n=page_n)
            self.total_tax = Amount({"value": total_tax}, value_key="value", page_n=page_n)
            self.total_excl = Amount({"value": total_excl}, value_key="value", page_n=page_n)

        # Invoke Document constructor
        super(Receipt, self).__init__(input_file)

        # Run checks
        self._checklist()

        # Reconstruct extra fields
        self._reconstruct()

    def __str__(self):
        return "-----Receipt data-----\n" \
               "Filename: %s\n" \
               "Total amount: %s \n" \
               "Date: %s\n" \
               "Category: %s\n" \
               "Time: %s\n" \
               "Merchant name: %s\n" \
               "Taxes: %s\n" \
               "Total taxes: %s\n" \
               "----------------------" % \
               (
                   self.filename,
                   self.total_incl.value,
                   self.date.value,
                   self.category.value,
                   self.time.value,
                   self.merchant_name.value,
                   " - ".join([str(t) for t in self.taxes]),
                   self.total_tax.value
               )

    def build_from_api_prediction(self, api_prediction, page_n=0):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param page_n: Page number for multi pages pdf input
        :return: (void) set the object attributes with api prediction values
        """
        self.locale = Locale(api_prediction["locale"], page_n=page_n)
        self.total_incl = Amount(api_prediction["total_incl"], value_key="value", page_n=page_n)
        self.date = Date(api_prediction["date"], value_key="value", page_n=page_n)
        self.category = Field(api_prediction["category"], page_n=page_n)
        self.merchant_name = Field(api_prediction["supplier"], value_key="value", page_n=page_n)
        self.time = Field(api_prediction["time"], value_key="value", page_n=page_n)
        self.taxes = [
            Tax(tax_prediction, page_n=page_n, value_key="value", rate_key="rate", code_key="code")
            for tax_prediction in api_prediction["taxes"]]
        self.orientation = Orientation(api_prediction["orientation"], page_n=page_n)
        self.total_tax = Amount({"value": None, "probability": 0.}, value_key="value", page_n=page_n)
        self.total_excl = Amount({"value": None, "probability": 0.}, value_key="value", page_n=page_n)

    @staticmethod
    def compare(receipt=None, ground_truth=None):
        """
        :param receipt: Receipt object to compare
        :param ground_truth: Ground truth Receipt object
        :return: Accuracy and precisions metrics
        """
        assert receipt is not None
        assert ground_truth is not None
        assert isinstance(receipt, Receipt)
        assert isinstance(ground_truth, Receipt)

        metrics = {}

        # Compute Accuracy metrics
        metrics.update(Receipt.compute_accuracy(receipt, ground_truth))

        # Compute precision metrics
        metrics.update(Receipt.compute_precision(receipt, ground_truth))

        return metrics

    @staticmethod
    def request(
            input_file,
            base_url,
            expense_receipt_token=None,
            version="3",
            include_words=False
    ):
        """
        Make request to expense_receipts endpoint
        :param input_file: Input object
        :param base_url: API base URL
        :param expense_receipt_token: Expense_receipts API token
        :param include_words: Include Mindee vision words in http_response
        :param version: API version
        """
        url = os.path.join(base_url, "expense_receipts", "v" + version, "predict")
        return request(url, input_file, expense_receipt_token, include_words)

    def _checklist(self):
        """
        Call check methods
        """
        self.checklist = {"taxes_match_total_incl": self.__taxes_match_total()}

    def _reconstruct(self):
        """
        Call fields reconstruction methods
        """
        self.__reconstruct_total_excl_from_tcc_and_taxes()
        self.__reconstruct_total_tax()

    # Checks
    def __taxes_match_total(self):
        """
        Check receipt rule of matching between taxes and total_incl
        :return: True if rule matches, False otherwise
        """
        # Check taxes and total amount exist
        if len(self.taxes) == 0 or self.total_incl.value is None:
            return False

        # Reconstruct total_incl from taxes
        total_vat = 0
        reconstructed_total = 0
        for tax in self.taxes:
            if tax.value is None or tax.rate is None:
                return False
            total_vat += tax.value
            reconstructed_total += tax.value + 100 * tax.value / tax.rate

        # Sanity check
        if total_vat <= 0:
            return False

        # Crate epsilon
        eps = 1 / (100 * total_vat)

        if self.total_incl.value * (1 - eps) - 0.02 <= reconstructed_total <= self.total_incl.value * (
                1 + eps) + 0.02:
            for tax in self.taxes:
                tax.probability = 1
            self.total_tax.probability = 1.
            self.total_incl.probability = 1.
            return True
        else:
            return False

    # Reconstruct
    def __reconstruct_total_excl_from_tcc_and_taxes(self):
        """
        Set self.total_excl with Amount object
        The total_excl Amount value is the difference between total_incl and sum of taxes
        The total_excl Amount probability is the product of self.taxes probabilities multiplied by total_incl probability
        """
        if len(self.taxes) and self.total_incl.value is not None:
            total_excl = {
                "value": self.total_incl.value - Field.array_sum(self.taxes),
                "probability": Field.array_probability(self.taxes) * self.total_incl.probability
            }
            self.total_excl = Amount(total_excl, value_key="value", reconstructed=True)

    def __reconstruct_total_tax(self):
        """
        Set self.total_tax with Amount object
        The total_tax Amount value is the sum of all self.taxes value
        The total_tax Amount probability is the product of self.taxes probabilities
        """
        if len(self.taxes) and self.total_tax.value is None:
            total_tax = {
                "value": sum([tax.value if tax.value is not None else 0 for tax in self.taxes]),
                "probability": Field.array_probability(self.taxes)
            }
            if total_tax["value"] > 0:
                self.total_tax = Amount(total_tax, value_key="value", reconstructed=True)

    @staticmethod
    def compute_accuracy(receipt, ground_truth):
        """
        :param receipt: Receipt object to compare
        :param ground_truth: Ground truth Receipt object
        :return: Accuracy metrics
        """
        return {
            "__acc__total_incl": ground_truth.total_incl == receipt.total_incl,
            "__acc__total_excl": ground_truth.total_excl == receipt.total_excl,
            "__acc__receipt_date": ground_truth.date == receipt.date,
            "__acc__total_tax": ground_truth.total_tax == receipt.total_tax,
            "__acc__taxes": Tax.compare_arrays(receipt.taxes, ground_truth.taxes),
        }

    @staticmethod
    def compute_precision(receipt, ground_truth):
        """
        :param receipt: Receipt object to compare
        :param ground_truth: Ground truth Receipt object
        :return: Precision metrics
        """
        precisions = {
            "__pre__total_incl": Benchmark.scalar_precision_score(
                receipt.total_incl, ground_truth.total_incl),
            "__pre__total_excl": Benchmark.scalar_precision_score(
                receipt.total_excl, ground_truth.total_excl),
            "__pre__receipt_date": Benchmark.scalar_precision_score(
                receipt.date, ground_truth.date),
            "__pre__total_tax": Benchmark.scalar_precision_score(
                receipt.total_tax, ground_truth.total_tax)}

        if len(receipt.taxes) == 0:
            precisions["__pre__taxes"] = None
        else:
            precisions["__pre__taxes"] = Tax.compare_arrays(receipt.taxes, ground_truth.taxes)

        return precisions



from mindee.fields import Field


class Locale(Field):
    def __init__(
            self,
            locale_prediction,
            value_key="value",
            reconstructed=False,
            page_n=None
    ):
        """
        :param locale_prediction: Locale prediction object from HTTP response
        :param value_key: Key to use in the locale_prediction dict
        :param reconstructed: Bool for reconstructed object (not extracted in the API)
        :param page_n: Page number for multi pages pdf
        """
        super(Locale, self).__init__(
            locale_prediction,
            value_key=value_key,
            reconstructed=reconstructed,
            page_n=page_n
        )

        if "language" not in locale_prediction.keys() or locale_prediction["language"] == "N/A":
            self.language = None
        else:
            self.language = locale_prediction["language"]

        if "country" not in locale_prediction.keys() or locale_prediction["country"] == "N/A":
            self.country = None
        else:
            self.country = locale_prediction["country"]

        if "currency" not in locale_prediction.keys() or locale_prediction["currency"] == "N/A":
            self.currency = None
        else:
            self.currency = locale_prediction["currency"]



import json
import pytest
from mindee import Passport


@pytest.fixture
def passport_object():
    json_repsonse = json.load(open("./tests/data/passport/v1/passport.json"))
    return Passport(json_repsonse["predictions"][0])


@pytest.fixture
def passport_object_from_scratch():
    return Passport(
        country="FR",
        id_number="sqd12354",
        birth_date="1989-10-19",
        expiry_date="2027-12-01",
        issuance_date="2017-11-31",
        birth_place="Paris",
        gender="M",
        surname="Doe",
        mrz1="P<GBRPUDARSAN<<HENERT<<<<<<<<<<<<<<<<<<<<<<<",
        mrz2="7077979792GBR9505209M1704224<<<<<<<<<<<<<<00",
        given_names=["John", "Jane"],
        mrz="P<GBRPUDARSAN<<HENERT<<<<<<<<<<<<<<<<<<<<<<<7077979792GBR9505209M1704224<<<<<<<<<<<<<<00",
        full_name="John Doe"
    )


@pytest.fixture
def passport_object_all_na():
    json_repsonse = json.load(open("./tests/data/passport/v1/passport_all_na.json"))
    return Passport(json_repsonse["predictions"][0])


def test_constructor(passport_object):
    assert passport_object.mrz1.value == "P<GBRPUDARSAN<<HENERT<<<<<<<<<<<<<<<<<<<<<<<"
    assert passport_object.mrz2.value == "7077979792GBR9505209M1704224<<<<<<<<<<<<<<00"
    assert passport_object.id_number.value == "707797979"
    assert type(passport_object.__str__()) == str
    assert passport_object.is_expired()
    assert passport_object.all_checks()


def test_all_na(passport_object_all_na):
    assert passport_object_all_na.mrz.value is None
    assert passport_object_all_na.country.value is None
    assert passport_object_all_na.id_number.value is None
    assert passport_object_all_na.birth_date.value is None
    assert passport_object_all_na.expiry_date.value is None
    assert passport_object_all_na.issuance_date.value is None
    assert passport_object_all_na.birth_place.value is None
    assert passport_object_all_na.gender.value is None
    assert passport_object_all_na.surname.value is None
    assert passport_object_all_na.mrz1.value is None
    assert passport_object_all_na.mrz2.value is None
    assert len(passport_object_all_na.given_names) == 0


def test_checklist(passport_object):
    for check in passport_object.checklist.values():
        assert check is True


def test_checklist_all_na(passport_object_all_na):
    for check in passport_object_all_na.checklist.values():
        assert check is False


def test_checksum():
    mrz = "7077979792GBR9505209M1704224<<<<<<<<<<<<<<00"
    assert Passport.check_sum(mrz[0:10] + mrz[13:20] + mrz[21:43]) == mrz[43]


def test_wrong_checksum():
    mrz = "7077974792GBR9505209M1704224<<<<<<<<<<<<<<00"
    assert Passport.check_sum(mrz[0:10] + mrz[13:20] + mrz[21:43]) != mrz[43]
    mrz = "7077974792GBR9505209M1404224<<<<<<<<<<<<<<00"
    assert Passport.check_sum(mrz[0:10] + mrz[13:20] + mrz[21:43]) != mrz[43]
    mrz = "7077974792GBR9505209M1404224<<<<<<<<<<<<<<08"
    assert Passport.check_sum(mrz[0:10] + mrz[13:20] + mrz[21:43]) != mrz[43]


def test_checksum_with_personal_number_alpha():
    mrz = "XDB0661884ESP9502138F1808122RE20050024133894"
    assert Passport.check_sum(mrz[28:42]) == mrz[42]


def test_compare_1(passport_object):
    # Compare same object must return all True
    benchmark = Passport.compare(passport_object, passport_object)
    for value in benchmark.values():
        assert value is True


def test_compare_3(passport_object, passport_object_all_na):
    # Compare full object and empty object
    benchmark = Passport.compare(passport_object, passport_object_all_na)
    for value in benchmark.values():
        assert value is False


def test_compare_5(passport_object_from_scratch):
    # Compare passport from class
    benchmark = Passport.compare(passport_object_from_scratch, passport_object_from_scratch)
    for key in benchmark.keys():
        if "__acc__" in key:
            assert benchmark[key] is True


def test_compare_6(passport_object_from_scratch):
    # Compare passport from class with empty given_names
    passport_object_from_scratch.given_names = []
    benchmark = Passport.compare(passport_object_from_scratch, passport_object_from_scratch)
    for key in benchmark.keys():
        if "__acc__" in key:
            assert benchmark[key] is True
        elif "__pre__" in key:
            assert benchmark[key] in [True, None]


def test_empty_object_works():
    passport = Passport()
    assert passport.full_name.value is None




import json
import pytest
from mindee import FinancialDocument


@pytest.fixture
def financial_doc_from_invoice_object():
    invoice_json_repsonse = json.load(open("./tests/data/invoices/v1/invoice.json"))
    return FinancialDocument(invoice_json_repsonse["predictions"][0])


@pytest.fixture
def financial_doc_from_receipt_object():
    receipt_json_repsonse = json.load(open("./tests/data/expense_receipts/v3/receipt.json"))
    return FinancialDocument(receipt_json_repsonse["predictions"][0])


@pytest.fixture
def financial_doc_object_from_scratch():
    return FinancialDocument(
        locale="fr",
        total_incl=12,
        total_excl=15,
        date="2018-12-21",
        invoice_number="001",
        due_date="2019-01-01",
        taxes={(1, 10), (2, 20)},
        merchant_name="Amazon",
        payment_details="1231456498799765",
        company_number="asdqsdae",
        vat_number="1231231232",
        orientation=0,
        total_tax=3,
        time="12:15"
    )


@pytest.fixture
def financial_doc_from_receipt_object_all_na():
    json_repsonse = json.load(open("./tests/data/expense_receipts/v3/receipt_all_na.json"))
    return FinancialDocument(json_repsonse["predictions"][0])


@pytest.fixture
def financial_doc_from_invoice_object_all_na():
    json_repsonse = json.load(open("./tests/data/invoices/v1/invoice_all_na.json"))
    return FinancialDocument(json_repsonse["predictions"][0])


@pytest.fixture
def receipt_pred():
    return json.load(open("./tests/data/expense_receipts/v3/receipt_all_na.json"))["predictions"][0]


@pytest.fixture
def invoice_pred():
    return json.load(open("./tests/data/invoices/v1/invoice_all_na.json"))["predictions"][0]


def test_constructor_1(financial_doc_from_invoice_object):
    assert financial_doc_from_invoice_object.date.value == "2020-02-17"


def test_constructor_2(financial_doc_from_receipt_object):
    assert financial_doc_from_receipt_object.date.value == "2016-02-26"


def test_all_na_receipt(financial_doc_from_receipt_object_all_na):
    assert financial_doc_from_receipt_object_all_na.orientation.value == 0
    assert financial_doc_from_receipt_object_all_na.locale.value is None
    assert financial_doc_from_receipt_object_all_na.total_incl.value is None
    assert financial_doc_from_receipt_object_all_na.date.value is None
    assert financial_doc_from_receipt_object_all_na.merchant_name.value is None
    assert financial_doc_from_receipt_object_all_na.total_tax.value is None
    assert len(financial_doc_from_receipt_object_all_na.taxes) == 0


def test_all_na_invoice(financial_doc_from_invoice_object_all_na):
    assert financial_doc_from_invoice_object_all_na.orientation.value == 0
    assert financial_doc_from_invoice_object_all_na.locale.value is None
    assert financial_doc_from_invoice_object_all_na.total_incl.value is None
    assert financial_doc_from_invoice_object_all_na.date.value is None
    assert financial_doc_from_invoice_object_all_na.merchant_name.value is None
    assert financial_doc_from_invoice_object_all_na.total_tax.value is None
    assert len(financial_doc_from_invoice_object_all_na.taxes) == 0


def test__str__invoice(financial_doc_from_invoice_object):
    assert type(financial_doc_from_invoice_object.__str__()) == str


def test__str__receipt(financial_doc_from_receipt_object):
    assert type(financial_doc_from_receipt_object.__str__()) == str


# Business tests from receipt
def test__receipt_reconstruct_total_excl_from_total_and_taxes_1(receipt_pred):
    # no incl implies no reconstruct for total excl
    receipt_pred["total_incl"] = {"value": "N/A", "probability": 0.}
    receipt_pred["taxes"] = [{"rate": 20, "value": 9.5, "probability": 0.9}]
    financial_doc = FinancialDocument(receipt_pred)
    assert financial_doc.total_excl.value is None


def test__receipt_reconstruct_total_excl_from_total_and_taxes_2(receipt_pred):
    # no taxes implies no reconstruct for total excl
    receipt_pred["total_incl"] = {"value": 12.54, "probability": 0.}
    receipt_pred["taxes"] = []
    financial_doc = FinancialDocument(receipt_pred)
    assert financial_doc.total_excl.value is None


def test__receipt_reconstruct_total_excl_from_total_and_taxes_3(receipt_pred):
    # working example
    receipt_pred["total_incl"] = {"value": 12.54, "probability": 0.5}
    receipt_pred["taxes"] = [{"rate": 20, "value": 0.5, "probability": 0.1},
                             {"rate": 10, "value": 4.25, "probability": 0.6}]
    financial_doc = FinancialDocument(receipt_pred)
    assert financial_doc.total_excl.probability == 0.03
    assert financial_doc.total_excl.value == 7.79


def test__receipt_reconstruct_total_tax_1(receipt_pred):
    # no taxes implies no reconstruct for total tax
    receipt_pred["taxes"] = []
    financial_doc = FinancialDocument(receipt_pred)
    assert financial_doc.total_tax.value is None


def test__receipt_reconstruct_total_tax_2(receipt_pred):
    # working example
    receipt_pred["taxes"] = [{"rate": 20, "value": 10.2, "probability": 0.5},
                             {"rate": 10, "value": 40.0, "probability": 0.1}]
    financial_doc = FinancialDocument(receipt_pred)
    assert financial_doc.total_tax.value == 50.2
    assert financial_doc.total_tax.probability == 0.05


def test__receipt_taxes_match_total_incl_1(receipt_pred):
    # matching example
    receipt_pred["total_incl"] = {"value": 507.25, "probability": 0.6}
    receipt_pred["taxes"] = [{"rate": 20, "value": 10.99, "probability": 0.5},
                             {"rate": 10, "value": 40.12, "probability": 0.1}]
    financial_doc = FinancialDocument(receipt_pred)
    assert financial_doc.checklist["taxes_match_total_incl"] is True
    assert financial_doc.total_incl.probability == 1.
    for tax in financial_doc.taxes:
        assert tax.probability == 1.


def test__receipt_taxes_match_total_incl_2(receipt_pred):
    # not matching example with close error
    receipt_pred["total_incl"] = {"value": 507.25, "probability": 0.6}
    receipt_pred["taxes"] = [{"rate": 20, "value": 10.9, "probability": 0.5},
                             {"rate": 10, "value": 40.12, "probability": 0.1}]
    financial_doc = FinancialDocument(receipt_pred)
    assert financial_doc.checklist["taxes_match_total_incl"] is False


def test__receipt_taxes_match_total_incl_3(receipt_pred):
    # sanity check with null tax
    receipt_pred["total_incl"] = {"value": 507.25, "probability": 0.6}
    receipt_pred["taxes"] = [{"rate": 20, "value": 0., "probability": 0.5}]
    financial_doc = FinancialDocument(receipt_pred)
    assert financial_doc.checklist["taxes_match_total_incl"] is False


# Business tests from invoice
def test__invoice_reconstruct_total_excl_from_total_and_taxes_1(invoice_pred):
    # no incl implies no reconstruct for total excl
    invoice_pred["total_incl"] = {"amount": "N/A", "probability": 0.}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 9.5, "probability": 0.9}]
    financial_doc = FinancialDocument(invoice_pred)
    assert financial_doc.total_excl.value is None


def test__invoice_reconstruct_total_excl_from_total_and_taxes_2(invoice_pred):
    # no taxes implies no reconstruct for total excl
    invoice_pred["total_incl"] = {"amount": 12.54, "probability": 0.}
    invoice_pred["taxes"] = []
    financial_doc = FinancialDocument(invoice_pred)
    assert financial_doc.total_excl.value is None


def test__invoice_reconstruct_total_excl_from_total_and_taxes_3(invoice_pred):
    # working example
    invoice_pred["total_incl"] = {"amount": 12.54, "probability": 0.5}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 0.5, "probability": 0.1},
                             {"rate": 10, "amount": 4.25, "probability": 0.6}]
    financial_doc = FinancialDocument(invoice_pred)
    assert financial_doc.total_excl.probability == 0.03
    assert financial_doc.total_excl.value == 7.79


def test__invoice_reconstruct_total_tax_1(invoice_pred):
    # no taxes implies no reconstruct for total tax
    invoice_pred["taxes"] = []
    financial_doc = FinancialDocument(invoice_pred)
    assert financial_doc.total_tax.value is None


def test__invoice_reconstruct_total_tax_2(invoice_pred):
    # working example
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.2, "probability": 0.5},
                             {"rate": 10, "amount": 40.0, "probability": 0.1}]
    financial_doc = FinancialDocument(invoice_pred)
    assert financial_doc.total_tax.value == 50.2
    assert financial_doc.total_tax.probability == 0.05


def test__invoice_taxes_match_total_incl_1(invoice_pred):
    # matching example
    invoice_pred["total_incl"] = {"amount": 507.25, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.99, "probability": 0.5},
                             {"rate": 10, "amount": 40.12, "probability": 0.1}]
    financial_doc = FinancialDocument(invoice_pred)
    assert financial_doc.checklist["taxes_match_total_incl"] is True
    assert financial_doc.total_incl.probability == 1.
    for tax in financial_doc.taxes:
        assert tax.probability == 1.


def test__invoice_taxes_match_total_incl_2(invoice_pred):
    # not matching example with close error
    invoice_pred["total_incl"] = {"amount": 507.25, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.9, "probability": 0.5},
                             {"rate": 10, "amount": 40.12, "probability": 0.1}]
    financial_doc = FinancialDocument(invoice_pred)
    assert financial_doc.checklist["taxes_match_total_incl"] is False


def test__invoice_taxes_match_total_incl_3(invoice_pred):
    # sanity check with null tax
    invoice_pred["total_incl"] = {"amount": 507.25, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 0., "probability": 0.5}]
    financial_doc = FinancialDocument(invoice_pred)
    assert financial_doc.checklist["taxes_match_total_incl"] is False


def test_compare_1(financial_doc_from_invoice_object):
    # Compare same object must return all True
    benchmark = FinancialDocument.compare(financial_doc_from_invoice_object, financial_doc_from_invoice_object)
    for value in benchmark.values():
        assert value is True


def test_compare_2(financial_doc_from_invoice_object, financial_doc_from_invoice_object_all_na):
    # Compare full object and empty object
    benchmark = FinancialDocument.compare(financial_doc_from_invoice_object, financial_doc_from_invoice_object_all_na)
    for key in set(benchmark.keys()) - {"time"}:
        assert benchmark[key] is False


def test_compare_3(financial_doc_object_from_scratch):
    # Compare financial doc from class
    benchmark = FinancialDocument.compare(
        financial_doc_object_from_scratch,
        financial_doc_object_from_scratch
    )
    for key in benchmark.keys():
        if "__acc__" in key:
            assert benchmark[key] is True


def test_compare_4(financial_doc_object_from_scratch):
    # Compare financial doc from class with empty taxes
    financial_doc_object_from_scratch.taxes = []
    benchmark = FinancialDocument.compare(
        financial_doc_object_from_scratch,
        financial_doc_object_from_scratch
    )
    for key in benchmark.keys():
        if "__acc__" in key:
            assert benchmark[key] is True
        elif "__pre__" in key:
            assert benchmark[key] in [True, None]


def test_empty_object_works():
    financial_doc = FinancialDocument()
    assert financial_doc.total_tax.value is None



import numpy as np
import matplotlib.pyplot as plt


def autolabel(ax, rects):
    """
    :param ax: Matplotlib ax
    :param rects: Matplotlib rectangles from chart's bars
    :return:
    """
    for rect in rects:
        height = rect.get_height()
        ax.annotate('{}'.format(height),
                    xy=(rect.get_x() + rect.get_width() / 2, height-height/10),
                    xytext=(0, 3), fontSize=5, color="#ffffff",
                    textcoords="offset points",
                    ha='center', va='bottom', rotation=90)


def plot_metrics(metrics, accuracies, precisions, save_path):
    """
    :param metrics: List of metrics names
    :param accuracies: List of accuracy values
    :param precisions: List of precision values
    :param save_path: Path to save the figure
    :return: (void) plot the precision and accuracy bar charts and save the figure in save_path
    """
    x = np.arange(len(metrics))  # the label locations
    width = 0.4  # the width of the bars

    fig, ax = plt.subplots()
    fig.subplots_adjust(bottom=0.15)
    rects1 = ax.bar(x - width / 2, accuracies, width, color='#fd3246', label='Accuracy')
    rects2 = ax.bar(x + width / 2, precisions, width, color='#007af9', label='Precision')

    autolabel(ax, rects1)
    autolabel(ax, rects2)
    # Add some text for labels, title and custom x-axis tick labels, etc.
    ax.set_ylabel('%')
    ax.set_title('Metrics')
    ax.set_xticks(x)
    ax.set_xticklabels(metrics, rotation=45, fontsize=6)
    ax.legend(loc='lower left')
    plt.grid(True, linestyle='--', color='#e1e1e1', alpha=0.4)

    plt.savefig(save_path, dpi=300)


import json

from mindee.http import HTTPException
from mindee.inputs import Inputs
from mindee.documents import Document
from mindee.documents.car_plate import CarPlate
from mindee.documents.receipt import Receipt
from mindee.documents.financial_document import FinancialDocument
from mindee.documents.invoice import Invoice
from mindee.documents.passport import Passport
from mindee.benchmark import Benchmark


class Client(object):
    def __init__(
            self,
            expense_receipt_token=None,
            invoice_token=None,
            passport_token=None,
            license_plate_token=None,
            raise_on_error=True
    ):
        """
        :param expense_receipt_token: expense_receipt Mindee API token, see https://mindee.com
        :param invoice_token: invoice Mindee API token, see https://mindee.com
        :param passport_token: passport Mindee API token, see https://mindee.com
        :param license_plate_token: license_plate Mindee API token, see https://mindee.com
        :param raise_on_error: (bool, default True) raise an Exception on HTTP errors
        """
        assert type(raise_on_error) == bool
        self.raise_on_error = raise_on_error
        self.base_url = "https://api.mindee.net/products/"
        self.expense_receipt_token = expense_receipt_token
        self.invoice_token = invoice_token
        self.passport_token = passport_token
        self.license_plate_token = license_plate_token

    def parse_receipt(
            self,
            file,
            input_type="path",
            version="3",
            cut_pdf=True,
            include_words=False
    ):
        """
        :param include_words: Bool, extract all words into http_response
        :param cut_pdf: Automatically reconstruct pdf with more than 4 pages
        :param input_type: String in {'path', 'stream', 'base64'}
        :param file: Receipt filepath (allowed jpg, png, tiff, pdf)
        :param version: expense_receipt api version
        :return: Wrapped response with Receipts objects parsed
        """
        if not self.expense_receipt_token:
            raise Exception("Missing 'expense_receipt_token' arg in parse_receipt() function.")

        input_file = Inputs(file, input_type, cut_pdf=cut_pdf)

        response = Receipt.request(
            input_file,
            self.base_url,
            self.expense_receipt_token,
            version,
            include_words
        )

        return self._wrap_response(input_file, response, "receipt")

    def _wrap_response(
            self,
            input_file,
            response,
            document_type
    ):
        """
        :param input_file: Input object
        :param response: HTTP response
        :param document_type: Document class in {"receipt", "invoice", "financial_document", "passport", "license_plate"}
        :return: Full response object
        """
        dict_response = response.json()
        if response.status_code != 200 and self.raise_on_error:
            raise HTTPException(
                "Receipt API %s HTTP error: %s" % (response.status_code, json.dumps(dict_response)))
        elif response.status_code != 200:
            return Response(
                http_response=dict_response,
                pages=[],
                document=None,
                document_type=document_type
            )

        return Response.format_response(dict_response, document_type, input_file)

    def parse_passport(
            self,
            file,
            input_type="path",
            version="1",
            cut_pdf=True
    ):
        """
        :param cut_pdf: Automatically reconstruct pdf with more than 4 pages
        :param input_type: String in {'path', 'stream', 'base64'}
        :param file: Passport filepath (allowed jpg, png, pdf)
        :param version: passport api version
        :return: Wrapped response with passports objects parsed
        """
        if not self.passport_token:
            raise Exception("Missing 'passport_token' arg in parse_passport() function.")

        input_file = Inputs(file, input_type, cut_pdf=cut_pdf)

        response = Passport.request(
            input_file,
            self.base_url,
            self.passport_token,
            version
        )

        return self._wrap_response(input_file, response, "passport")

    def parse_license_plate(
            self,
            file,
            input_type="path",
            version="1",
            cut_pdf=True
    ):
        """
        :param cut_pdf: Automatically reconstruct pdf with more than 4 pages
        :param input_type: String in {'path', 'stream', 'base64'}
        :param file: CarPlate filepath (allowed jpg, png, pdf)
        :param version: license_plates api version
        :return: Wrapped response with CarPlates objects parsed
        """
        if not self.license_plate_token:
            raise Exception("Missing 'license_plate_token' arg in license_plate_token() function.")

        input_file = Inputs(file, input_type, cut_pdf=cut_pdf)

        response = CarPlate.request(
            input_file,
            self.base_url,
            self.license_plate_token,
            version
        )

        return self._wrap_response(input_file, response, "license_plate")

    def parse_invoice(
            self,
            file,
            input_type="path",
            version="1",
            cut_pdf=True,
            include_words=False
    ):
        """
        :param include_words: Bool, extract all words into http_response
        :param cut_pdf: Automatically reconstruct pdf with more than 4 pages
        :param input_type: String in {'path', 'stream', 'base64'}
        :param file: Invoice filepath (allowed jpg, png, pdf)
        :param version: invoices api version
        :return: Wrapped response with Invoices objects parsed
        """
        if not self.invoice_token:
            raise Exception("Missing 'invoice_token' arg in parse_invoice() function.")

        input_file = Inputs(file, input_type, cut_pdf=cut_pdf)

        response = Invoice.request(
            input_file,
            self.base_url,
            self.invoice_token,
            version,
            include_words
        )

        return self._wrap_response(input_file, response, "invoice")

    def parse_financial_document(
            self,
            file,
            input_type="path",
            cut_pdf=True,
            include_words=False
    ):
        """
        :param include_words: Bool, extract all words into http_response
        :param cut_pdf: Automatically reconstruct pdf with more than 4 pages
        :param input_type: String in {'path', 'stream', 'base64'}
        :param file: Invoice or Receipt filepath (allowed jpg, png, pdf)
        :return: Wrapped response with FinancialDocument objects parsed
        """
        if not self.invoice_token or not self.expense_receipt_token:
            raise Exception("parse_invoice() function must include 'invoice_token' and 'expense_receipt_token' args.")

        input_file = Inputs(file, input_type, cut_pdf=cut_pdf)

        response = FinancialDocument.request(
            input_file,
            self.base_url,
            self.expense_receipt_token,
            self.invoice_token,
            include_words
        )

        return self._wrap_response(input_file, response, "financial_document")


class Response(object):
    def __init__(
            self,
            http_response=None,
            pages=None,
            document=None,
            document_type=None
    ):
        """
        :param http_response: HTTP response object
        :param pages: List of document objects
        :param document: reconstructed object from all pages
        :param document_type: Document class in {"receipt", "invoice", "financial_document", "passport", "license_plate"}
        """
        self.http_response = http_response
        self.document_type = document_type
        if document_type == "receipt":
            self.receipt = document
            self.receipts = pages
        if document_type == "invoice":
            self.invoice = document
            self.invoices = pages
        if document_type == "financial_document":
            self.financial_document = document
            self.financial_documents = pages
        if document_type == "passport":
            self.passport = document
            self.passports = pages
        if document_type == "license_plate":
            self.license_plate = document
            self.license_plates = pages

    def dump(self, path):
        """
        :param path: file path for storing the response object
        :return: (void) save the json response
        """
        with open(path, "w") as fp:
            json.dump(self.http_response, fp)

    @staticmethod
    def load(json_path):
        """
        :param json_path: file of the json reponse to restore
        :return: Full response object loaded from json file
        """
        try:
            with open(json_path) as fp:
                json_response = json.load(fp)

            file_input = Inputs.load(
                json_response["input_type"],
                json_response["filename"],
                json_response["filepath"],
                json_response["file_extension"]
            )
            response = Response.format_response(
                json_response,
                document_type=json_response["document_type"],
                input_file=file_input
            )
            return response
        except:
            raise Exception("Unable to load file.")

    @staticmethod
    def format_response(json_response, document_type, input_file):
        """
        :param input_file: Input object
        :param json_response: json response from HTTP call
        :param document_type: Document class in {"receipt", "invoice", "financial_document", "passport", "license_plate"}
        :return: Full Response object
        """
        json_response["document_type"] = document_type
        json_response["input_type"] = input_file.input_type
        json_response["filename"] = input_file.filename
        json_response["filepath"] = input_file.filepath
        json_response["file_extension"] = input_file.file_extension
        pages = []
        for page_n, page_prediction in enumerate(json_response["predictions"]):
            if document_type == "receipt":
                pages.append(
                    Receipt(
                        api_prediction=page_prediction,
                        input_file=input_file,
                        page_n=page_n
                    )
                )
            elif document_type == "invoice":
                pages.append(
                    Invoice(
                        api_prediction=page_prediction,
                        input_file=input_file,
                        page_n=page_n
                    )
                )
            elif document_type == "financial_document":
                pages.append(
                    FinancialDocument(
                        api_prediction=page_prediction,
                        input_file=input_file,
                        page_n=page_n
                    )
                )
            elif document_type == "passport":
                pages.append(
                    Passport(
                        api_prediction=page_prediction,
                        input_file=input_file,
                        page_n=page_n
                    )
                )
            elif document_type == "license_plate":
                pages.append(
                    CarPlate(
                        api_prediction=page_prediction,
                        input_file=input_file,
                        page_n=page_n
                    )
                )
            else:
                raise Exception("Document type not supported.")

        document = Document.merge_pages(pages)

        return Response(
            http_response=json_response,
            pages=pages,
            document=document,
            document_type=document_type
        )



import datetime
import pytz
from mindee.fields import Field

ISO8601_DATE_FORMAT = '%Y-%m-%d'
ISO8601_DATETIME_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


class Date(Field):
    def __init__(
            self,
            date_prediction,
            value_key="iso",
            reconstructed=False,
            page_n=None
    ):
        """
        :param date_prediction: Date prediction object from HTTP response
        :param value_key: Key to use in the date_prediction dict
        :param reconstructed: Bool for reconstructed object (not extracted in the API)
        :param page_n: Page number for multi pages pdf
        """
        super(Date, self).__init__(
            date_prediction,
            value_key=value_key,
            reconstructed=reconstructed,
            page_n=page_n
        )

        try:
            self.date_object = datetime.datetime. \
                strptime(self.value, ISO8601_DATE_FORMAT). \
                replace(tzinfo=pytz.utc).date()
        except:
            self.date_object = None
            self.probability = 0.
            self.value = None



import pytest
import json
from mindee import Response, Inputs
from mindee.documents import Document


@pytest.fixture
def dummy_file_input():
    file_input = Inputs("./tests/data/expense_receipts/receipt.jpg")
    return file_input


def test_constructor():
    document = Document()
    assert document.checklist == {}
    with pytest.raises(NotImplementedError):
        document.request()
    with pytest.raises(NotImplementedError):
        document._checklist()


# Plate tests

def test_responseWrapper_plate(dummy_file_input):
    response = json.load(open('./tests/data/license_plates/v1/plate.json'))
    parsed_plate = Response.format_response(response, "license_plate", dummy_file_input)
    assert parsed_plate.license_plate.license_plates[0].value == "7EQE707"


# Invoice tests

def test_responseWrapper_invoice(dummy_file_input):
    response = json.load(open('./tests/data/invoices/v1/invoice.json'))
    parsed_invoice = Response.format_response(response, "invoice", dummy_file_input)
    assert parsed_invoice.invoice.invoice_date.value == "2020-02-17"


from mindee import Invoice

json_repsonse = json.load(open("./tests/data/invoices/v1/invoice.json"))
invoice_0 = Invoice(json_repsonse["predictions"][0])
invoice_1 = Invoice(json_repsonse["predictions"][1])


def test_merge_pages():
    merge_document = Document.merge_pages([invoice_0, invoice_1])
    assert merge_document.invoice_date.value == "2020-02-17"


# Receipt tests

def test_responseWrapper_receipt(dummy_file_input):
    response = json.load(open('./tests/data/expense_receipts/v3/receipt.json'))
    parsed_receipt = Response.format_response(response, "receipt", dummy_file_input)
    assert parsed_receipt.receipt.date.value == "2016-02-26"


# Financial document tests

def test_responseWrapper_financial_document_with_receipt(dummy_file_input):
    response = json.load(open('./tests/data/expense_receipts/v3/receipt.json'))
    parsed_financial_doc = Response.format_response(response, "financial_document", dummy_file_input)
    assert parsed_financial_doc.financial_document.date.value == "2016-02-26"


def test_responseWrapper_financial_document_with_invoice(dummy_file_input):
    response = json.load(open('./tests/data/invoices/v1/invoice.json'))
    parsed_financial_doc = Response.format_response(response, "financial_document", dummy_file_input)
    assert parsed_financial_doc.financial_document.date.value == "2020-02-17"




from mindee.benchmark import Benchmark
from mindee.documents import Document
from mindee.fields import Field
from mindee.fields.date import Date
from datetime import datetime
from mindee.http import request
import os


class Passport(Document):
    def __init__(
            self,
            api_prediction=None,
            input_file=None,
            country=None,
            id_number=None,
            birth_date=None,
            expiry_date=None,
            issuance_date=None,
            birth_place=None,
            gender=None,
            surname=None,
            mrz1=None,
            mrz2=None,
            given_names=None,
            mrz=None,
            full_name=None,
            page_n=0
    ):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param input_file: Input object
        :param country: country value for creating Passport object from scratch
        :param id_number: id_number value for creating Passport object from scratch
        :param birth_date: birth_date value for creating Passport object from scratch
        :param expiry_date: expiry_date value for creating Passport object from scratch
        :param issuance_date: issuance_date value for creating Passport object from scratch
        :param birth_place: birth_place value for creating Passport object from scratch
        :param gender: gender value for creating Passport object from scratch
        :param surname: surname value for creating Passport object from scratch
        :param mrz1: mrz1 value for creating Passport object from scratch
        :param mrz2: mrz2 value for creating Passport object from scratch
        :param given_names: given_names value for creating Passport object from scratch
        :param mrz: mrz value for creating Passport object from scratch
        :param full_name: full_name value for creating Passport object from scratch
        :param page_n: Page number for multi pages pdf input
        """
        # Raw data
        self.type = "Passport"
        self.country = None
        self.id_number = None
        self.birth_date = None
        self.expiry_date = None
        self.issuance_date = None
        self.birth_place = None
        self.gender = None
        self.surname = None
        self.mrz1 = None
        self.mrz2 = None
        self.given_names = []
        self.mrz = None
        self.full_name = None

        if api_prediction is not None:
            self.build_from_api_prediction(api_prediction)
        else:
            self.country = Field({"value": country}, value_key="value", page_n=page_n)
            self.id_number = Field({"value": id_number}, value_key="value", page_n=page_n)
            self.birth_date = Date({"value": birth_date}, value_key="value", page_n=page_n)
            self.expiry_date = Date({"value": expiry_date}, value_key="value", page_n=page_n)
            self.issuance_date = Date({"value": issuance_date}, value_key="value", page_n=page_n)
            self.birth_place = Field({"value": birth_place}, value_key="value", page_n=page_n)
            self.gender = Field({"value": gender}, value_key="value", page_n=page_n)
            self.surname = Field({"value": surname}, value_key="value", page_n=page_n)
            self.mrz1 = Field({"value": mrz1}, value_key="value", page_n=page_n)
            self.mrz2 = Field({"value": mrz2}, value_key="value", page_n=page_n)
            if given_names is not None:
                self.given_names = [Field({"value": g}, value_key="value", page_n=page_n) for g in given_names]
            self.mrz = Field({"value": mrz}, value_key="value", page_n=page_n)
            self.full_name = Field({"value": full_name}, value_key="value", page_n=page_n)

        # Invoke Document constructor
        super(Passport, self).__init__(input_file)

        # Run checks
        self._checklist()

        # Reconstruct extra fields
        self._reconstruct()

    def build_from_api_prediction(self, api_prediction, page_n=0):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param page_n: Page number for multi pages pdf input
        :return: (void) set the object attributes with api prediction values
        """
        self.country = Field(api_prediction["country"], page_n=page_n)
        self.id_number = Field(api_prediction["id_number"], page_n=page_n)
        self.birth_date = Date(api_prediction["birth_date"], "value", page_n=page_n)
        self.expiry_date = Date(api_prediction["expiry_date"], "value", page_n=page_n)
        self.issuance_date = Date(api_prediction["issuance_date"], "value", page_n=page_n)
        self.birth_place = Field(api_prediction["birth_place"], page_n=page_n)
        self.gender = Field(api_prediction["gender"], page_n=page_n)
        self.surname = Field(api_prediction["surname"], page_n=page_n)
        self.mrz1 = Field(api_prediction["mrz1"], page_n=page_n)
        self.mrz2 = Field(api_prediction["mrz2"], page_n=page_n)
        self.given_names = [Field(given_name, page_n=page_n) for given_name in api_prediction["given_names"]]
        self.mrz = Field({"value": None, "probability": 0.}, page_n=page_n)
        self.full_name = Field({"value": None, "probability": 0.}, page_n=page_n)

    def __str__(self):
        return "-----Passport data-----\n" \
               "Filename: %s \n" \
               "Full name: %s \n" \
               "Given names: %s \n" \
               "Surname: %s\n" \
               "Country: %s\n" \
               "ID Number: %s\n" \
               "Issuance date: %s\n" \
               "Birth date: %s\n" \
               "Expiry date: %s\n" \
               "MRZ 1: %s\n" \
               "MRZ 2: %s\n" \
               "MRZ: %s\n" \
               "----------------------" % \
               (
                   self.filename,
                   self.full_name.value,
                   " ".join(
                       [given_name.value if given_name.value is not None else "" for given_name in self.given_names]),
                   self.surname.value,
                   self.country.value,
                   self.id_number.value,
                   self.issuance_date.value,
                   self.birth_date.value,
                   self.expiry_date.value,
                   self.mrz1.value,
                   self.mrz2.value,
                   self.mrz.value
               )

    @staticmethod
    def compare(passport=None, ground_truth=None):
        """
        :param passport: Passport object to compare
        :param ground_truth: Ground truth Passport object
        :return: Accuracy and precisions metrics
        """
        assert passport is not None
        assert ground_truth is not None
        assert isinstance(passport, Passport)
        assert isinstance(ground_truth, Passport)

        metrics = {}

        # Compute Accuracy metrics
        metrics.update(Passport.compute_accuracy(passport, ground_truth))

        # Compute precision metrics
        metrics.update(Passport.compute_precision(passport, ground_truth))

        return metrics

    def is_expired(self):
        """
        :return: True if the passport is expired, False otherwise
        """
        return self.expiry_date.date_object < datetime.date(datetime.now())

    @staticmethod
    def request(input_file, base_url, passport_token=None, version="1"):
        """
        Make request to passport endpoint
        """
        url = os.path.join(base_url, "passport", "v" + version, "predict")
        return request(url, input_file, passport_token)

    def _reconstruct(self):
        """
        Call fields reconstruction methods
        """
        self.__reconstruct_mrz()
        self.__reconstruct_full_name()

    def _checklist(self, *args):
        """
        Call check methods
        """
        self.checklist = {
            "mrz_id_number_checksum": self.__mrz_id_number_checksum(),
            "mrz_date_of_birth_checksum": self.__mrz_date_of_birth_checksum(),
            "mrz_expiration_date_checksum": self.__mrz_expiration_date_checksum(),
            "mrz_personal_number_checksum": self.__mrz_personal_number_checksum(),
            "mrz_last_name_checksum": self.__mrz_last_name_checksum(),
            "mrz_checksum": self.__mrz__checksum()
        }

    # Checks
    def __mrz__checksum(self):
        """
        :return: True if the all MRZ checksums are validated, False otherwise
        """
        return self.__mrz_id_number_checksum() and \
               self.__mrz_date_of_birth_checksum() and \
               self.__mrz_expiration_date_checksum() and \
               self.__mrz_personal_number_checksum() and \
               self.__mrz_last_name_checksum()

    def __mrz_id_number_checksum(self):
        """
        :return: True if id number MRZ checksum is validated, False otherwise
        """
        if self.mrz2.value is None:
            return False
        if Passport.check_sum(self.mrz2.value[:9]) == self.mrz2.value[9]:
            self.id_number.probability = 1.
            return True

    def __mrz_date_of_birth_checksum(self):
        """
        :return: True if date of birth MRZ checksum is validated, False otherwise
        """
        if self.mrz2.value is None:
            return False
        if Passport.check_sum(self.mrz2.value[13:19]) == self.mrz2.value[19]:
            self.birth_date.probability = 1.
            return True

    def __mrz_expiration_date_checksum(self):
        """
        :return: True if expiry date MRZ checksum is validated, False otherwise
        """
        if self.mrz2.value is None:
            return False
        if Passport.check_sum(self.mrz2.value[21:27]) == self.mrz2.value[27]:
            self.expiry_date.probability = 1.
            return True

    def __mrz_personal_number_checksum(self):
        """
        :return: True if personal number MRZ checksum is validated, False otherwise
        """
        if self.mrz2.value is None:
            return False
        return Passport.check_sum(self.mrz2.value[28:42]) == self.mrz2.value[42]

    def __mrz_last_name_checksum(self):
        """
        :return: True if last name MRZ checksum is validated, False otherwise
        """
        if self.mrz2.value is None:
            return False
        if Passport.check_sum(self.mrz2.value[0:10] + self.mrz2.value[13:20] + self.mrz2.value[21:43]) == \
                self.mrz2.value[43]:
            self.surname.probability = 1.
            return True

    @staticmethod
    def check_sum(s):
        """
        https://en.wikipedia.org/wiki/Machine-readable_passport
        :param s: string
        :return: checksum value for string s
        """
        checker = 0
        alpha_to_num = {c: 10 + i for i, c in enumerate('ABCDEFGHIJKLMNOPQRSTUVWXYZ')}
        for i, c in enumerate(s):
            if i % 3 == 0:
                weight = 7
            elif i % 3 == 1:
                weight = 3
            else:
                weight = 1

            if c == '<':
                val = 0
            elif c.isalpha():
                val = alpha_to_num[c]
            else:
                val = int(c)
            checker += val * weight
        return str(checker % 10)

    # Reconstruct
    def __reconstruct_mrz(self):
        """
        Set self.mrz with Field object
        The mrz Field value is the concatenation of mrz1 and mr2
        The mrz Field probability is the product of mrz1 and mrz2 probabilities
        """
        if self.mrz1.value is not None \
                and self.mrz2.value is not None \
                and self.mrz.value is None:
            mrz = {
                "value": self.mrz1.value + self.mrz2.value,
                "probability": Field.array_probability([self.mrz1.probability, self.mrz2.probability])
            }
            self.mrz = Field(mrz, reconstructed=True)

    def __reconstruct_full_name(self):
        """
        Set self.full_name with Field object
        The full_name Field value is the concatenation of first given name and last name
        The full_name Field probability is the product of first given name and last name probabilities
        """
        if self.surname.value is not None \
                and len(self.given_names) != 0 \
                and self.given_names[0].value is not None \
                and self.full_name.value is None:
            full_name = {
                "value": self.given_names[0].value + " " + self.surname.value,
                "probability": Field.array_probability([self.surname.probability, self.given_names[0].probability])
            }
            self.full_name = Field(full_name, reconstructed=True)

    @staticmethod
    def compute_accuracy(passport, ground_truth):
        """
        :param passport: Passport object to compare
        :param ground_truth: Ground truth Passport object
        :return: Accuracy metrics
        """
        return {
            "__acc__country": ground_truth.country == passport.country,
            "__acc__id_number": ground_truth.id_number == passport.id_number,
            "__acc__birth_date": ground_truth.birth_date == passport.birth_date,
            "__acc__expiry_date": ground_truth.expiry_date == passport.expiry_date,
            "__acc__issuance_date": ground_truth.issuance_date == passport.issuance_date,
            "__acc__gender": ground_truth.gender == passport.gender,
            "__acc__surname": ground_truth.surname == passport.surname,
            "__acc__mrz1": ground_truth.mrz1 == passport.mrz1,
            "__acc__mrz2": ground_truth.mrz2 == passport.mrz2,
            "__acc__given_names": Field.compare_arrays(passport.given_names, ground_truth.given_names),
            "__acc__mrz": ground_truth.mrz == passport.mrz,
            "__acc__full_name": ground_truth.full_name == passport.full_name,
        }

    @staticmethod
    def compute_precision(passport, ground_truth):
        """
        :param passport: Passport object to compare
        :param ground_truth: Ground truth Passport object
        :return: Precision metrics
        """
        precisions = {
            "__pre__country": Benchmark.scalar_precision_score(passport.country, ground_truth.country),
            "__pre__id_number": Benchmark.scalar_precision_score(passport.id_number, ground_truth.id_number),
            "__pre__birth_date": Benchmark.scalar_precision_score(passport.birth_date, ground_truth.birth_date),
            "__pre__expiry_date": Benchmark.scalar_precision_score(passport.expiry_date, ground_truth.expiry_date),
            "__pre__issuance_date": Benchmark.scalar_precision_score(passport.issuance_date,
                                                                     ground_truth.issuance_date),
            "__pre__gender": Benchmark.scalar_precision_score(passport.gender, ground_truth.gender),
            "__pre__surname": Benchmark.scalar_precision_score(passport.surname, ground_truth.surname),
            "__pre__mrz1": Benchmark.scalar_precision_score(passport.mrz1, ground_truth.mrz1),
            "__pre__mrz2": Benchmark.scalar_precision_score(passport.mrz2, ground_truth.mrz2),
            "__pre__mrz": Benchmark.scalar_precision_score(passport.mrz, ground_truth.mrz),
            "__pre__full_name": Benchmark.scalar_precision_score(passport.full_name, ground_truth.full_name),
        }

        if len(passport.given_names) == 0:
            precisions["__pre__given_names"] = None
        else:
            precisions["__pre__given_names"] = Field.compare_arrays(passport.given_names, ground_truth.given_names)

        return precisions



import copy


class Document(object):
    def __init__(self, input_file=None):
        self.filepath = None
        self.filename = None
        self.file_extension = None

        if input_file is not None:
            self.filepath = input_file.filepath
            self.filename = input_file.filename
            self.file_extension = input_file.file_extension
        self.checklist = {}

    def request(self, *args):
        raise NotImplementedError()

    def _checklist(self, *args):
        raise NotImplementedError()

    def _reconstruct(self, *args):
        pass

    def all_checks(self):
        return all(self.checklist)

    @staticmethod
    def merge_pages(page_documents):
        """
        :param page_documents: Document object list
        :return: A single Document where each field is set with the maximum probability field
        """
        document = copy.deepcopy(page_documents[0])
        attributes = [a for a in dir(document)]
        for doc in page_documents:
            for attribute in attributes:
                if not hasattr(getattr(doc, attribute), "probability"):
                    continue

                if getattr(doc, attribute).probability > getattr(document, attribute).probability:
                    setattr(document, attribute, getattr(doc, attribute))

        return document



import datetime
from mindee.fields.date import Date


def test_constructor():
    field_dict = {
        'iso': "2018-04-01",
        'probability': 0.1,
        'segmentation': {
            "bounding_box": [
                [0.016, 0.707],
                [0.414, 0.707],
                [0.414, 0.831],
                [0.016, 0.831]
            ]
        }
    }
    date = Date(field_dict)
    assert date.value == "2018-04-01"


def test_constructor_no_date():
    field_dict = {
        'iso': "N/A",
        'probability': 0.1
    }
    date = Date(field_dict)
    assert date.value is None




from mindee.benchmark import Benchmark

from mindee.fields.amount import Amount
from mindee.fields.date import Date
from mindee.fields.locale import Locale
from mindee.fields.orientation import Orientation
from mindee.fields.tax import Tax
from mindee.documents import Document
from mindee.fields import Field
from mindee.http import request
from mindee.documents.invoice import Invoice
from mindee.documents.receipt import Receipt
import os


class FinancialDocument(Document):
    def __init__(
            self,
            api_prediction=None,
            input_file=None,
            locale=None,
            total_incl=None,
            total_excl=None,
            date=None,
            invoice_number=None,
            due_date=None,
            taxes=None,
            merchant_name=None,
            payment_details=None,
            company_number=None,
            vat_number=None,
            orientation=None,
            total_tax=None,
            time=None,
            page_n=0
    ):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param input_file: Input object
        :param locale: locale value for creating FinancialDocument object from scratch
        :param total_incl: total_incl value for creating FinancialDocument object from scratch
        :param total_excl: total_excl value for creating FinancialDocument object from scratch
        :param date: date value for creating FinancialDocument object from scratch
        :param invoice_number: invoice_number value for creating FinancialDocument object from scratch
        :param due_date: due_date value for creating FinancialDocument object from scratch
        :param taxes: taxes value for creating FinancialDocument object from scratch
        :param merchant_name: merchant_name value for creating FinancialDocument object from scratch
        :param payment_details: payment_details value for creating FinancialDocument object from scratch
        :param company_number: company_number value for creating FinancialDocument object from scratch
        :param vat_number: vat_number value for creating FinancialDocument object from scratch
        :param orientation: orientation value for creating FinancialDocument object from scratch
        :param total_tax: total_tax value for creating FinancialDocument object from scratch
        :param time: time value for creating FinancialDocument object from scratch
        :param page_n: Page number for multi pages pdf input
        """
        self.type = "FinancialDocument"
        self.locale = None
        self.total_incl = None
        self.total_excl = None
        self.date = None
        self.invoice_number = None
        self.due_date = None
        self.taxes = []
        self.merchant_name = None
        self.payment_details = None
        self.company_number = None
        self.vat_number = None
        self.orientation = None
        self.total_tax = None
        self.time = None

        if api_prediction is not None:
            self.build_from_api_prediction(api_prediction, input_file, page_n=page_n)
        else:
            self.orientation = Orientation({"value": orientation}, value_key="value", page_n=page_n)
            self.locale = Locale({"value": locale}, value_key="value", page_n=page_n)
            self.date = Date({"value": date}, value_key="value", page_n=page_n)
            self.due_date = Date({"value": due_date}, value_key="value", page_n=page_n)
            if taxes is not None:
                self.taxes = [
                    Tax({"value": t[0], "rate": t[1]}, page_n=page_n, value_key="value", rate_key="rate")
                    for t in taxes]
            self.total_incl = Amount({"value": total_incl}, value_key="value", page_n=page_n)
            self.total_excl = Amount({"value": total_excl}, value_key="value", page_n=page_n)
            self.merchant_name = Field({"value": merchant_name}, value_key="value", page_n=page_n)
            self.time = Field({"value": time}, value_key="value", page_n=page_n)
            self.total_tax = Amount({"value": total_tax}, value_key="value", page_n=page_n)
            self.vat_number = Field({"value": vat_number}, value_key="value", page_n=page_n)
            self.invoice_number = Field({"value": invoice_number}, value_key="value", page_n=page_n)
            self.payment_details = Field({"value": payment_details}, value_key="value", page_n=page_n)
            self.company_number = Field({"value": company_number}, value_key="value", page_n=page_n)

        # Invoke Document constructor
        super(FinancialDocument, self).__init__(input_file)

        # Run checks
        self._checklist()

        # Reconstruct extra fields
        self._reconstruct()

    def build_from_api_prediction(self, api_prediction, input_file, page_n=0):
        """
        :param api_prediction: Raw prediction from HTTP response
        :param input_file: Input object
        :param page_n: Page number for multi pages pdf input
        :return: (void) set the object attributes with api prediction values
        """
        if "invoice_number" in api_prediction.keys():
            invoice = Invoice(api_prediction, input_file, page_n=page_n)
            self.locale = invoice.locale
            self.total_incl = invoice.total_incl
            self.total_excl = invoice.total_excl
            self.date = invoice.invoice_date
            self.invoice_number = invoice.invoice_number
            self.due_date = invoice.due_date
            self.taxes = invoice.taxes
            self.merchant_name = invoice.supplier
            self.payment_details = invoice.payment_details
            self.company_number = invoice.company_number
            self.orientation = invoice.orientation
            self.total_tax = invoice.total_tax
            self.vat_number = invoice.vat_number
            self.time = Field({"value": None, "probability": 0.})
        else:
            receipt = Receipt(api_prediction, input_file, page_n=page_n)
            self.orientation = receipt.orientation
            self.date = receipt.date
            self.due_date = receipt.date
            self.taxes = receipt.taxes
            self.locale = receipt.locale
            self.total_incl = receipt.total_incl
            self.total_excl = receipt.total_excl
            self.merchant_name = receipt.merchant_name
            self.time = receipt.time
            self.total_tax = receipt.total_tax
            self.vat_number = Field({"value": None, "probability": 0.})
            self.invoice_number = Field({"value": None, "probability": 0.})
            self.payment_details = Field({"value": None, "probability": 0.})
            self.company_number = Field({"value": None, "probability": 0.})

    def __str__(self):
        return "-----Financial document-----\n" \
               "Filename: %s \n" \
               "Total amount: %s \n" \
               "Date: %s\n" \
               "Merchant name: %s\n" \
               "Total taxes: %s\n" \
               "----------------------" % \
               (
                   self.filename,
                   self.total_incl.value,
                   self.date.value,
                   self.merchant_name.value,
                   self.total_tax.value
               )

    @staticmethod
    def compare(financial_document=None, ground_truth=None):
        """
        :param financial_document: FinancialDocument object to compare
        :param ground_truth: Ground truth FinancialDocument object
        :return: Accuracy and precisions metrics
        """
        assert financial_document is not None
        assert ground_truth is not None
        assert isinstance(financial_document, FinancialDocument)
        assert isinstance(ground_truth, FinancialDocument)

        metrics = {}

        # Compute Accuracy metrics
        metrics.update(FinancialDocument.compute_accuracy(financial_document, ground_truth))

        # Compute precision metrics
        metrics.update(FinancialDocument.compute_precision(financial_document, ground_truth))

        return metrics

    @staticmethod
    def request(
            input_file,
            base_url,
            expense_receipt_token=None,
            invoice_token=None,
            include_words=False
    ):
        """
        Make request to invoices endpoint if .pdf, expense_receipts otherwise
        :param include_words: Bool, extract all words into http_response
        :param input_file: Input object
        :param base_url: API base URL
        :param expense_receipt_token: Expense receipts API token
        :param invoice_token: Invoices API token
        """
        if "pdf" in input_file.file_extension:
            url = os.path.join(base_url, "invoices", "v1", "predict")
            return request(url, input_file, invoice_token, include_words)
        else:
            url = os.path.join(base_url, "expense_receipts", "v3", "predict")
            return request(url, input_file, expense_receipt_token, include_words)

    def _checklist(self):
        """
        :return: Set of validation rules
        """
        self.checklist = {"taxes_match_total_incl": self.__taxes_match_total_incl()}

    # Checks
    def __taxes_match_total_incl(self):
        """
        Check invoice rule of matching between taxes and total_incl
        :return: True if rule matches, False otherwise
        """
        # Check taxes and total_incl exist
        if len(self.taxes) == 0 or self.total_incl.value is None:
            return False

        # Reconstruct total_incl from taxes
        total_vat = 0
        reconstructed_total = 0
        for tax in self.taxes:
            total_vat += tax.value
            reconstructed_total += tax.value + 100 * tax.value / tax.rate

        # Sanity check
        if total_vat <= 0:
            return False

        # Crate epsilon
        eps = 1 / (100 * total_vat)

        if self.total_incl.value * (1 - eps) - 0.02 <= reconstructed_total <= self.total_incl.value * (
                1 + eps) + 0.02:
            for tax in self.taxes:
                tax.probability = 1
            self.total_tax.probability = 1.
            self.total_incl.probability = 1.
            return True
        else:
            return False

    @staticmethod
    def compute_accuracy(financial_document, ground_truth):
        """
        :param financial_document: FinancialDocument object to compare
        :param ground_truth: Ground truth FinancialDocument object
        :return: Accuracy metrics
        """
        return {
            "__acc__total_incl": ground_truth.total_incl == financial_document.total_incl,
            "__acc__total_excl": ground_truth.total_excl == financial_document.total_excl,
            "__acc__invoice_date": ground_truth.date == financial_document.date,
            "__acc__invoice_number": ground_truth.invoice_number == financial_document.invoice_number,
            "__acc__due_date": ground_truth.due_date == financial_document.due_date,
            "__acc__total_tax": ground_truth.total_tax == financial_document.total_tax,
            "__acc__taxes": Tax.compare_arrays(financial_document.taxes, ground_truth.taxes),
        }

    @staticmethod
    def compute_precision(financial_document, ground_truth):
        """
        :param financial_document: FinancialDocument object to compare
        :param ground_truth: Ground truth FinancialDocument object
        :return: Precision metrics
        """
        precisions = {
            "__pre__total_incl": Benchmark.scalar_precision_score(
                financial_document.total_incl, ground_truth.total_incl),
            "__pre__total_excl": Benchmark.scalar_precision_score(
                financial_document.total_excl, ground_truth.total_excl),
            "__pre__invoice_date": Benchmark.scalar_precision_score(
                financial_document.date, ground_truth.date),
            "__pre__invoice_number": Benchmark.scalar_precision_score(
                financial_document.invoice_number, ground_truth.invoice_number),
            "__pre__due_date": Benchmark.scalar_precision_score(
                financial_document.due_date, ground_truth.due_date),
            "__pre__total_tax": Benchmark.scalar_precision_score(
                financial_document.total_tax, ground_truth.total_tax)}

        if len(financial_document.taxes) == 0:
            precisions["__pre__taxes"] = None
        else:
            precisions["__pre__taxes"] = Tax.compare_arrays(financial_document.taxes, ground_truth.taxes)

        return precisions



from mindee.fields.tax import Tax


def test_constructor():
    field_dict = {
        'value': "2",
        'rate': 0.2,
        'code': "QST",
        'probability': 0.1,
        'segmentation': {
            "bounding_box": [
                [0.016, 0.707],
                [0.414, 0.707],
                [0.414, 0.831],
                [0.016, 0.831]
            ]
        }
    }
    tax = Tax(field_dict, value_key="value")
    assert tax.value == 2
    assert tax.probability == 0.1
    assert tax.rate == 0.2
    assert len(tax.bbox) > 0
    assert type(str(tax)) == str


def test_constructor_no_rate():
    field_dict = {
        'value': "2",
        'rate': "AA",
        'probability': 0.1
    }
    tax = Tax(field_dict)
    assert tax.rate is None
    assert len(tax.bbox) == 0


def test_constructor_no_amount():
    field_dict = {
        'value': "NA",
        'rate': "AA",
        'code': "N/A",
        'probability': 0.1
    }
    tax = Tax(field_dict)
    assert tax.value is None
    assert type(str(tax)) == str



import statistics
import calendar
import json
import time
import os
from os import path
from mindee.plots import plot_metrics


class Benchmark(object):
    def __init__(
            self,
            output_dir,
            benchmark_name=None
    ):
        """
        :param output_dir: Directory path for saving benchmark results
        :param benchmark_name: Optional name of the benchmark folder inside output_dir
        """
        if not path.exists(output_dir):
            raise Exception(
                'Output directory %s does not exist. Please create it before running benchmark.' % output_dir)
        self.output_dir = output_dir
        self.results = None

        if benchmark_name is None:
            benchmark_name = str(calendar.timegm(time.gmtime()))

        self.benchmark_dir = os.path.join(self.output_dir, benchmark_name)
        os.mkdir(self.benchmark_dir)

        self.results_dir = os.path.join(self.output_dir, benchmark_name, "results")
        os.mkdir(self.results_dir)

    def initialize(self, result):
        """
        :param result: output from Document.compare method
        :return: (void) initialize benchmark results
        """
        self.results = {"file_ids": []}
        for key in result.keys():
            self.results[key] = []

    def add(self, result, file_id):
        """
        :param result: output from Document.compare method
        :param file_id: Unique file identifier
        :return: (void) add a data point to the benchmark
        """
        if self.results is None:
            self.initialize(result)

        assert set(result.keys()) == set(self.results.keys()) - {"file_ids"}
        for key in result.keys():
            if '__pre__' in key and result[key] is not None:
                self.results[key].append(result[key])
            if '__acc__' in key:
                self.results[key].append(result[key])

        self.results["file_ids"].append(file_id)

        with open(os.path.join(self.results_dir, '%s.json' % file_id), "w") as fp:
            json.dump(result, fp)

    def save(self):
        """
        :return: (void) plot precision and accuracy bar chart into output_dir folder
        """
        with open(os.path.join(self.benchmark_dir, 'results.json'), "w") as fp:
            json.dump(self.results, fp)

        metrics = []
        accuracies = []
        precisions = []
        for key in set(self.results.keys()) - {"file_ids"}:
            if "__acc__" in key:
                metrics.append(key[7:])
                accuracies.append(round(statistics.mean(self.results[key]) * 100, 2))
                precisions.append(round(statistics.mean(self.results['__pre__'+key[7:]]) * 100, 2))

        plot_metrics(metrics, accuracies, precisions, os.path.join(self.benchmark_dir, 'metrics.png'))

    @staticmethod
    def scalar_precision_score(field, ground_truth):
        """
        :param field: Field to compare
        :param ground_truth: Ground truth field to make the comparison on
        :return: True or False if the Field is set, None otherwise
        """
        if field.value is None:
            return None
        else:
            return field == ground_truth



from mindee.fields.locale import Locale


def test_constructor():
    field_dict = {
        "value": "en-EN",
        "language": "en",
        "country": "uk",
        "currency": "GBP",
        "probability": 0.1
    }
    locale = Locale(field_dict)
    assert locale.value == "en-EN"
    assert locale.language == "en"
    assert locale.country == "uk"
    assert locale.currency == "GBP"


def test_constructor_almost_empty_field():
    field_dict = {
        "value": "en-EN",
        "probability": 0.1
    }
    locale = Locale(field_dict)
    assert locale.language is None
    assert locale.country is None
    assert locale.currency is None



class Field(object):
    def __init__(
            self,
            abstract_prediction,
            value_key="value",
            reconstructed=False,
            extra_fields=None,
            page_n=None
    ):
        """
        :param abstract_prediction: Prediction object from HTTP response
        :param value_key: Key to use in the abstract_prediction dict
        :param reconstructed: Bool for reconstructed object (not extracted in the API)
        :param page_n: Page number for multi pages pdf
        :param extra_fields: extra field to get from the abstract_prediction and to set as attribute of the Field
        """
        self.page_n = page_n
        if value_key not in abstract_prediction or abstract_prediction[value_key] == "N/A":
            self.value = None
            self.probability = 0.
            self.bbox = []
        else:
            self.value = abstract_prediction[value_key]

            if "probability" in abstract_prediction:
                self.probability = abstract_prediction["probability"]
            else:
                self.probability = 0.

            if "segmentation" in abstract_prediction:
                self.bbox = abstract_prediction["segmentation"]["bounding_box"]
            else:
                self.bbox = []

        if extra_fields:
            for field_name in extra_fields:
                setattr(self, field_name, abstract_prediction[field_name])

        self.reconstructed = reconstructed

    def __eq__(self, other):
        if self.value is None and other.value is None:
            return True
        elif self.value is None or other.value is None:
            return False
        else:
            if type(self.value) == str:
                return self.value.lower() == other.value.lower()
            else:
                return self.value == other.value

    @staticmethod
    def compare_arrays(array1, array2, attr="value"):
        """
        :param array1: Array of Fields
        :param array2: Array of Fields
        :param attr: Attribute to compare
        :return: True if all elements in array1 exist in array2, False otherwise
        """
        set1 = set([getattr(f1, attr) for f1 in array1])
        set2 = set([getattr(f2, attr) for f2 in array2])
        return set1 == set2

    @staticmethod
    def array_probability(array):
        """
        :param array: Array of fields
        :return: Product of all the Fields probability in the array
        """
        product = 1
        for field in array:
            try:
                product *= field.probability
            except:
                return 0.
        return product

    @staticmethod
    def array_sum(array):
        """
        :param array: Array of fields
        :return: Sum of all the Fields values in the array
        """
        array_sum = 0
        for field in array:
            try:
                array_sum += field.value
            except:
                return 0.
        return array_sum




import re
from setuptools import find_packages, setup

__version__ = None


with open("README.md", "r", newline="", encoding="utf-8") as fh:
    long_description = fh.read()


APP_NAME = "mindee"
PACKAGE_NAME = "mindee"
GIT_URL = "https://github.com/publicMindee/mindee-api-python"


def make_requirements_list(file="requirements.txt", only_regular=True):
    """
    Make a list of package requirements from a requirements.txt file
    :param file: path to txt file
    :param only_regular: remove rows with /, #, space or empty
    :return:
    """

    with open(file) as f:
        lines = f.read().splitlines()
    if only_regular:
        regex = (
            "\/$|^#|^$$|^git\+"
        )  # remove line with /, starting by # or space or empty
        return [line for line in lines if not re.findall(regex, line)]
    else:
        return lines


setup(
    python_requires=">=3.0",
    name=f"{PACKAGE_NAME}",
    description="Mindee API helper library for python",
    version="v1.0.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=GIT_URL,
    packages=find_packages(),
    author="Mindee",
    author_email="contact@mindee.com",
    install_requires=make_requirements_list(),
    include_package_data=True
)


from mindee.fields.amount import Amount


def test_constructor():
    field_dict = {
        'amount': "2",
        'probability': 0.1,
        'segmentation': {
            "bounding_box": [
                [0.016, 0.707],
                [0.414, 0.707],
                [0.414, 0.831],
                [0.016, 0.831]
            ]
        }
    }
    amount = Amount(field_dict)
    assert amount.value == 2
    assert amount.value-amount.value == 0


def test_constructor_no_amount():
    field_dict = {
        'amount': "N/A",
        'probability': 0.1
    }
    amount = Amount(field_dict)
    assert amount.value is None



import requests


def request(url, input_file, token, include_words=False):
    """
    :param input_file: Input object
    :param url: Endpoint url
    :param token: X-Inferuser-Token
    :param include_words: Include Mindee vision words in http_response
    :return: requests response
    """
    input_file.file_object.seek(0)

    files = {"file": input_file.file_object.read()}

    headers = {"X-Inferuser-Token": token}

    params = {}
    if include_words:
        params["include_mvision"] = "true"

    response = requests.post(
        url+"?include_mvision=True",
        files=files,
        headers=headers,
        data=params
    )

    input_file.file_object.close()

    return response


class HTTPException(Exception):
    def __init__(self, message):
        self.message = message



import json
import pytest
from mindee import Invoice


@pytest.fixture
def invoice_object():
    json_repsonse = json.load(open("./tests/data/invoices/v1/invoice.json"))
    return Invoice(json_repsonse["predictions"][0])


@pytest.fixture
def invoice_object_from_scratch():
    return Invoice(
        locale="fr",
        total_incl=12,
        total_excl=15,
        invoice_date="2018-12-21",
        invoice_number="001",
        due_date="2019-01-01",
        taxes={(1, 10), (2, 20)},
        supplier="Amazon",
        payment_details="1231456498799765",
        company_number="asdqsdae",
        vat_number="1231231232",
        orientation=0,
        total_tax=3
    )


@pytest.fixture
def invoice_object_all_na():
    json_repsonse = json.load(open("./tests/data/invoices/v1/invoice_all_na.json"))
    return Invoice(json_repsonse["predictions"][0])


@pytest.fixture
def invoice_pred():
    return json.load(open("./tests/data/invoices/v1/invoice_all_na.json"))["predictions"][0]


# Technical tests
def test_constructor(invoice_object):
    assert invoice_object.invoice_date.value == "2020-02-17"
    assert invoice_object.checklist["taxes_match_total_incl"] is True
    assert invoice_object.checklist["taxes_match_total_excl"] is True
    assert invoice_object.checklist["taxes_plus_total_excl_match_total_incl"] is True
    assert invoice_object.total_tax.value == 97.98
    assert type(invoice_object.__str__()) == str


def test_all_na(invoice_object_all_na):
    assert invoice_object_all_na.locale.value is None
    assert invoice_object_all_na.total_incl.value is None
    assert invoice_object_all_na.total_excl.value is None
    assert invoice_object_all_na.invoice_date.value is None
    assert invoice_object_all_na.invoice_number.value is None
    assert invoice_object_all_na.due_date.value is None
    assert len(invoice_object_all_na.taxes) == 0
    assert invoice_object_all_na.supplier.value is None
    assert invoice_object_all_na.payment_details.value is None
    assert invoice_object_all_na.company_number.value is None
    assert invoice_object_all_na.orientation.value == 0


def test_checklist_on_empty(invoice_object_all_na):
    for check in invoice_object_all_na.checklist.values():
        assert check is False


# Business tests
def test__reconstruct_total_incl_from_taxes_plus_excl_1(invoice_pred):
    # no taxes implies no reconstruct for total incl
    invoice_pred["total_incl"] = {"amount": "N/A", "probability": 0.}
    invoice_pred["total_excl"] = {"amount": 240.5, "probability": 0.9}
    invoice_pred["taxes"] = []
    invoice = Invoice(invoice_pred)
    assert invoice.total_incl.value is None


def test__reconstruct_total_incl_from_taxes_plus_excl_2(invoice_pred):
    # no excl implies no reconstruct for total incl
    invoice_pred["total_incl"] = {"amount": "N/A", "probability": 0.}
    invoice_pred["total_excl"] = {"amount": "N/A", "probability": 0.}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 9.5, "probability": 0.9}]
    invoice = Invoice(invoice_pred)
    assert invoice.total_incl.value is None


def test__reconstruct_total_incl_from_taxes_plus_excl_3(invoice_pred):
    # incl already exists implies no reconstruct
    invoice_pred["total_incl"] = {"amount": 260, "probability": 0.4}
    invoice_pred["total_excl"] = {"amount": 240.5, "probability": 0.9}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 9.5, "probability": 0.9}]
    invoice = Invoice(invoice_pred)
    assert invoice.total_incl.value == 260
    assert invoice.total_incl.probability == 0.4


def test__reconstruct_total_incl_from_taxes_plus_excl_4(invoice_pred):
    # working example
    invoice_pred["total_incl"] = {"amount": "N/A", "probability": 0.}
    invoice_pred["total_excl"] = {"amount": 240.5, "probability": 0.9}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 9.5, "probability": 0.9}]
    invoice = Invoice(invoice_pred)
    assert invoice.total_incl.value == 250
    assert invoice.total_incl.probability == 0.81


def test__reconstruct_total_excl_from_tcc_and_taxes_1(invoice_pred):
    # no incl implies no reconstruct for total excl
    invoice_pred["total_incl"] = {"amount": "N/A", "probability": 0.}
    invoice_pred["total_excl"] = {"amount": "N/A", "probability": 0.}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 9.5, "probability": 0.9}]
    invoice = Invoice(invoice_pred)
    assert invoice.total_excl.value is None


def test__reconstruct_total_excl_from_tcc_and_taxes_2(invoice_pred):
    # no taxes implies no reconstruct for total excl
    invoice_pred["total_incl"] = {"amount": 1150.20, "probability": 0.7}
    invoice_pred["total_excl"] = {"amount": "N/A", "probability": 0.}
    invoice_pred["taxes"] = []
    invoice = Invoice(invoice_pred)
    assert invoice.total_excl.value is None


def test__reconstruct_total_excl_from_tcc_and_taxes_3(invoice_pred):
    # excl already exists implies no reconstruct
    invoice_pred["total_incl"] = {"amount": 1150.20, "probability": 0.7}
    invoice_pred["total_excl"] = {"amount": 1050.0, "probability": 0.4}
    invoice_pred["taxes"] = []
    invoice = Invoice(invoice_pred)
    assert invoice.total_excl.value == 1050.0
    assert invoice.total_excl.probability == 0.4


def test__reconstruct_total_excl_from_tcc_and_taxes_4(invoice_pred):
    # working example
    invoice_pred["total_incl"] = {"amount": 1150.20, "probability": 0.6}
    invoice_pred["total_excl"] = {"amount": "N/A", "probability": 0.}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.2, "probability": 0.5},
                             {"rate": 10, "amount": 40.0, "probability": 0.1}]
    invoice = Invoice(invoice_pred)
    assert invoice.total_excl.value == 1100
    assert invoice.total_excl.probability == 0.03


def test__reconstruct_total_tax_1(invoice_pred):
    # no taxes implies no reconstruct for total tax
    invoice_pred["taxes"] = []
    invoice = Invoice(invoice_pred)
    assert invoice.total_tax.value is None


def test__reconstruct_total_tax_2(invoice_pred):
    # working example
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.2, "probability": 0.5},
                             {"rate": 10, "amount": 40.0, "probability": 0.1}]
    invoice = Invoice(invoice_pred)
    assert invoice.total_tax.value == 50.2
    assert invoice.total_tax.probability == 0.05


def test__taxes_match_total_incl_1(invoice_pred):
    # matching example
    invoice_pred["total_incl"] = {"amount": 507.25, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.99, "probability": 0.5},
                             {"rate": 10, "amount": 40.12, "probability": 0.1}]
    invoice = Invoice(invoice_pred)
    assert invoice.checklist["taxes_match_total_incl"] is True
    assert invoice.total_incl.probability == 1.
    for tax in invoice.taxes:
        assert tax.probability == 1.


def test__taxes_match_total_incl_2(invoice_pred):
    # not matching example with close error
    invoice_pred["total_incl"] = {"amount": 507.25, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.9, "probability": 0.5},
                             {"rate": 10, "amount": 40.12, "probability": 0.1}]
    invoice = Invoice(invoice_pred)
    assert invoice.checklist["taxes_match_total_incl"] is False


def test__taxes_match_total_incl_3(invoice_pred):
    # sanity check with null tax
    invoice_pred["total_incl"] = {"amount": 507.25, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 0., "probability": 0.5}]
    invoice = Invoice(invoice_pred)
    assert invoice.checklist["taxes_match_total_incl"] is False


def test__taxes_match_total_excl_1(invoice_pred):
    # matching example
    invoice_pred["total_excl"] = {"amount": 456.15, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.99, "probability": 0.5},
                             {"rate": 10, "amount": 40.12, "probability": 0.1}]
    invoice = Invoice(invoice_pred)
    assert invoice.checklist["taxes_match_total_excl"] is True
    assert invoice.total_excl.probability == 1.
    for tax in invoice.taxes:
        assert tax.probability == 1.


def test__taxes_match_total_excl_2(invoice_pred):
    # not matching example  with close error
    invoice_pred["total_excl"] = {"amount": 456.15, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.9, "probability": 0.5},
                             {"rate": 10, "amount": 40.12, "probability": 0.1}]
    invoice = Invoice(invoice_pred)
    assert invoice.checklist["taxes_match_total_excl"] is False


def test__taxes_match_total_excl_3(invoice_pred):
    # sanity check with null tax
    invoice_pred["total_excl"] = {"amount": 507.25, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 0., "probability": 0.5}]
    invoice = Invoice(invoice_pred)
    assert invoice.checklist["taxes_match_total_incl"] is False


def test__taxes_plus_total_excl_match_total_incl_1(invoice_pred):
    # matching example
    invoice_pred["total_incl"] = {"amount": 507.25, "probability": 0.6}
    invoice_pred["total_excl"] = {"amount": 456.15, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.99, "probability": 0.5},
                             {"rate": 10, "amount": 40.12, "probability": 0.1}]
    invoice = Invoice(invoice_pred)
    assert invoice.checklist["taxes_plus_total_excl_match_total_incl"] is True
    assert invoice.total_incl.probability == 1.
    assert invoice.total_excl.probability == 1.
    for tax in invoice.taxes:
        assert tax.probability == 1.


def test__taxes_plus_total_excl_match_total_incl_2(invoice_pred):
    # not matching example
    invoice_pred["total_incl"] = {"amount": 507.2, "probability": 0.6}
    invoice_pred["total_excl"] = {"amount": 456.15, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 10.99, "probability": 0.5},
                             {"rate": 10, "amount": 40.12, "probability": 0.1}]
    invoice = Invoice(invoice_pred)
    assert invoice.checklist["taxes_plus_total_excl_match_total_incl"] is False


def test__taxes_plus_total_excl_match_total_incl_3(invoice_pred):
    # sanity check with null tax
    invoice_pred["total_excl"] = {"amount": 456.15, "probability": 0.6}
    invoice_pred["total_incl"] = {"amount": 507.25, "probability": 0.6}
    invoice_pred["taxes"] = [{"rate": 20, "amount": 0., "probability": 0.5}]
    invoice = Invoice(invoice_pred)
    assert invoice.checklist["taxes_match_total_incl"] is False


def test_compare_1(invoice_object):
    # Compare same object must return all True
    benchmark = Invoice.compare(invoice_object, invoice_object)
    for value in benchmark.values():
        assert value is True


def test_compare_2(invoice_object, invoice_object_all_na):
    # Compare full object and empty object
    benchmark = Invoice.compare(invoice_object, invoice_object_all_na)
    for key in benchmark.keys():
        assert benchmark[key] is False


def test_compare_3(invoice_object_from_scratch):
    # Compare invoice from class
    benchmark = Invoice.compare(invoice_object_from_scratch, invoice_object_from_scratch)
    for key in benchmark.keys():
        if "__acc__" in key:
            assert benchmark[key] is True


def test_compare_4(invoice_object_from_scratch):
    # Compare invoice from class with empty taxes
    invoice_object_from_scratch.taxes = []
    benchmark = Invoice.compare(invoice_object_from_scratch, invoice_object_from_scratch)
    for key in benchmark.keys():
        if "__acc__" in key:
            assert benchmark[key] is True
        elif "__pre__" in key:
            assert benchmark[key] in [True, None]


def test_empty_object_works():
    invoice = Invoice()
    assert invoice.total_tax.value is None

